import json
import ast

if __name__ == '__main__':
    # test_string = "{'spotlight': 'TRUE'}"
    # test_string = "{'spotlight': True}"
    test_string = "{'spotlight': True}"
    # converted = json.loads(test_string)
    # converted = ast.literal_eval(test_string)
    # converted = eval(test_string.replace("TRUE", "True"))
    converted = eval(test_string)

    print(converted)



from pathlib import Path
from sparc_me.core.utils import convert_schema_excel_to_json


if __name__ == "__main__":
    source_path = Path(
        r"C:\Users\clin864\OneDrive - The University of Auckland\Desktop\sparc-codathon\sparc-me\sparc_me\resources\templates\version_2_0_0\schema.xlsx")
    dest_path = Path(
        r"C:\Users\clin864\OneDrive - The University of Auckland\Desktop\sparc-codathon\sparc-me\sparc_me\resources\templates\version_2_0_0\schema.json")

    convert_schema_excel_to_json(source_path=source_path, dest_path=dest_path)





import os
import shutil
import json
import tempfile

from pathlib import Path
from distutils.dir_util import copy_tree
from typing import Dict

import pandas as pd
from styleframe import StyleFrame
from xlrd import XLRDError
from datetime import datetime, timezone
from sparc_me.core.utils import check_row_exist, get_sub_folder_paths_in_folder, validate_metadata_file
from sparc_me.core.metadata import Metadata, Sample, Subject


class Dataset(object):
    def __init__(self):
        DEFAULT_DATASET_VERSION = "2.0.0"
        EXTENSIONS = [".xlsx"]

        self._template_version = DEFAULT_DATASET_VERSION
        self._version = DEFAULT_DATASET_VERSION
        self._current_path = Path(__file__).parent.resolve()
        self._resources_path = Path.joinpath(self._current_path, "../resources")
        self._template_dir = Path()
        self._template = dict()
        self._subjects = {}

        self._dataset_path = Path()
        self._dataset = dict()
        self._metadata_extensions = EXTENSIONS
        self._column_based = ["dataset_description", "code_description"]
        self._subject_id_field = None
        self._sample_id_field = None
        self._metadata: Dict[str, Metadata] = {}

    def set_path(self, path):
        """
        Set the dataset path

        :param path: path to the dataset directory
        :type path: string
        """
        self.set_dataset_path(path)

    def set_dataset_path(self, path):
        """
        Set the path to the dataset

        :param path: path to the dataset directory
        :type path: string
        """
        self._dataset_path = Path(path)
        Sample._dataset_path = self._dataset_path
        Subject._dataset_path = self._dataset_path

    def get_dataset_path(self):
        """
        Return the path to the dataset directory
        :return: path to the dataset directory
        :rtype: string
        """
        return str(self._dataset_path)

    def get_dataset(self):
        """
        :return: current dataset dict
        """
        return self._dataset

    def _get_template_dir(self, version):
        """
        Get template directory path

        :return: path to the template dataset
        :rtype: Path
        """
        version = "version_" + version
        template_dir = self._resources_path / "templates" / version / "DatasetTemplate"

        return template_dir

    def _set_template_version(self, version):
        """
        Choose a template version

        :param version: template version
        :type version: string
        """
        version = self._convert_version_format(version)
        self._template_version = version
        self._set_version_specific_variables(version)

    def _set_version_specific_variables(self, version):
        """Set version specific variables

        :param version: SDS version to use Ex: 2_0_0
        :type version: string
        :raises ValueError: if the given version is not an acceptable SDS version
        """
        if version == "2_0_0":
            self._subject_id_field = "subject id"
            self._sample_id_field = "sample id"
        elif version == "1_2_3":
            self._subject_id_field = "subject_id"
            self._sample_id_field = "sample_id"
        else:
            error_msg = f"Unsupported version {version}"
            raise ValueError(error_msg)

    def _load(self, dir_path):
        """
        Load the input dataset into a dictionary

        :param dir_path: path to the dataset dictionary
        :type dir_path: string
        :return: loaded dataset
        :rtype: dict
        """
        dataset = dict()

        dir_path = Path(dir_path)
        for path in dir_path.iterdir():
            if path.suffix in self._metadata_extensions:
                try:
                    metadata = pd.read_excel(path)
                except XLRDError:
                    metadata = pd.read_excel(path, engine='openpyxl')

                metadata = metadata.dropna(how="all")
                metadata = metadata.loc[:, ~metadata.columns.str.contains('^Unnamed')]

                key = path.stem
                value = {
                    "path": path,
                    "metadata": metadata
                }
            else:
                key = path.name
                value = path

            dataset[key] = value

        return dataset

    def create_empty_dataset(self, version='2.0.0'):
        """
        Create an empty dataset from template via dataset version
        :param version: the dataset version
        :type version: '2.0.0' | '1.2.3'
        """
        self.load_from_template(version=version)

    def load_from_template(self, version):
        """
        Load dataset from SPARC template

        :param version: template version
        :type version: string
        :return: loaded dataset
        :rtype: dict
        """
        self._set_version(version)
        # self._dataset_path = self._get_template_dir(self._version)
        template_dataset_path = self._get_template_dir(self._version)
        self._dataset = self._load(str(template_dataset_path))

        self._generate_metadata()

    def _convert_version_format(self, version):
        """
        Convert version format
        :param version: dataset/template version
        :type version: string
        :return: version in the converted format
        :rtype:
        """
        version = version.replace(".", "_")

        if "_" not in version:
            version = version + "_0_0"

        return version

    def _set_version(self, version):
        """
        Set dataset version version

        :param version: dataset version
        :type version: string
        """
        version = self._convert_version_format(version)

        self._version = version
        self._set_version_specific_variables(version)

    def _load_template(self, version):
        """
        Load template

        :param version: template version
        :type version: string
        :return: loaded template
        :rtype: dict
        """

        version = self._convert_version_format(version)
        self._set_template_version(version)
        self._template_dir = self._get_template_dir(self._template_version)
        self._template = self._load(str(self._template_dir))

        return self._template

    def _save_template(self, save_dir, version=None):
        """
        Save the template directory locally
        TODO: will delete later

        :param save_dir: path to the output folder
        :type save_dir: string
        :param version: template version
        :type version: string
        """
        if version:
            version = self._convert_version_format(version)
            template_dir = self._get_template_dir(version)
        elif not version and self._template_version:
            template_dir = self._get_template_dir(self._template_version)
        else:
            raise ValueError("Template path not found.")

        copy_tree(str(template_dir), str(save_dir))

    def load_dataset(self, dataset_path=None, from_template=False, version=None):
        """
        Load the input dataset into a dictionary

        :param dataset_path: path to the dataset
        :type dataset_path: string
        :param from_template: whether to load the dataset from a SPARC template
        :type from_template: bool
        :param version: dataset version
        :type version: string
        :return: loaded dataset
        :rtype: dict
        """
        if version:
            self._set_version(version)

        if not self._dataset_path:
            self._dataset_path = Path(dataset_path)

        if from_template:
            self._dataset = self.load_from_template(version=version)
        else:
            self._dataset = self._load(dataset_path)
            self._generate_metadata()

        return self._dataset

    def save(self, save_dir="", remove_empty=False, keep_style=False):
        """
        Save dataset

        :param save_dir: path to the dest dir
        :type save_dir: string
        :param remove_empty: (optional) If True, remove rows which do not have values in the "Value" field
        :type remove_empty: bool
        """
        if not self._dataset:
            msg = "Dataset not defined. Please load the dataset or the template dataset in advance."
            raise ValueError(msg)
        if save_dir == "":
            save_dir = self._dataset_path
        save_dir = Path(save_dir)
        save_dir.mkdir(parents=True, exist_ok=True)

        for key, value in self._dataset.items():
            if isinstance(value, dict):
                file_path = Path(value.get("path"))
                filename = file_path.name
                data = value.get("metadata")

                if remove_empty:
                    data = self._filter(data, filename)

                if isinstance(data, pd.DataFrame):
                    self._set_version(self._version)
                    template_dir = self._get_template_dir(self._version)

                    if keep_style:
                        sf = StyleFrame.read_excel_as_template(str(template_dir / filename), data)
                        writer = StyleFrame.ExcelWriter(Path.joinpath(save_dir, filename))
                        sf.to_excel(writer)
                        writer.save()
                    else:
                        data.to_excel(Path.joinpath(save_dir, filename), index=False)

            elif Path(value).is_dir():
                dir_name = Path(value).name
                dir_path = Path.joinpath(save_dir, dir_name)
                copy_tree(str(value), str(dir_path), update=1)

            elif Path(value).is_file():
                filename = Path(value).name
                file_path = Path.joinpath(save_dir, filename)
                try:
                    shutil.copyfile(value, file_path)
                except shutil.SameFileError:
                    # overwrite file by copy, remove then rename
                    file_path_tmp = str(file_path) + "_tmp"
                    shutil.copyfile(value, file_path_tmp)
                    os.remove(file_path)
                    os.rename(file_path_tmp, file_path)

        for gitkeep_path in save_dir.rglob('.gitkeep'):
            gitkeep_path.unlink()

    def load_metadata(self, path):
        """
        Load & update a single metadata

        :param path: path to the metadata file
        :type path: string
        :return: metadata
        :rtype: Pandas.DataFrame
        """
        path = Path(path)
        try:
            metadata = pd.read_excel(path)
        except XLRDError:
            metadata = pd.read_excel(path, engine='openpyxl')

        filename = path.stem
        self._dataset[filename] = {
            "path": path,
            "metadata": metadata
        }

        return metadata

    def _filter(self, metadata, filename):
        """
        Remove column/row if values not set

        :param metadata: metadata
        :type metadata: Pandas.DataFrame
        :param filename: name of the metadata
        :type filename: string
        :return: updated metadata
        :rtype: Pandas.DataFrame
        """
        if "dataset_description" in filename:
            # For the dataset_description metadata, remove rows which do not have values in the "Value" fields
            metadata = metadata.dropna(subset=["Value"])

        return metadata

    def list_metadata_files(self, version, print_list=True):
        """
        list all metadata_files based on the metadata files in the template dataset

        :param version: reference template version
        :type version: string
        :return: all metadata metadata_files
        :rtype: list
        """
        metadata_files = list()

        self._load_template(version=version)

        for key, value in self._template.items():
            if isinstance(value, dict):
                file_path = Path(value.get("path"))
                metadata_file = file_path.stem
                metadata_files.append(metadata_file)

        if print_list:
            print("metadata_files:")
            for metadata_file in metadata_files:
                print(metadata_file)

        return metadata_files

    def list_elements(self, metadata_file, axis=0, version=None):
        """
        List field from a metadata file

        :param metadata_file: metadata metadata_file
        :type metadata_file: string
        :param axis: If axis=0, column-based. list all column headers. i.e. the first row.
                     If axis=1, row-based. list all row index. i.e. the first column in each row
        :type axis: int
        :param version: reference template version
        :type version: string
        :return: a list of fields
        :rtype: list
        """
        fields = None
        metadata_file = validate_metadata_file(metadata_file, version)
        if metadata_file == "dataset_description":
            axis = 1

        if version:
            version = self._convert_version_format(version)
            template_dir = self._get_template_dir(version)

            element_description_file = template_dir / "../schema.xlsx"

            try:
                element_description = pd.read_excel(element_description_file, sheet_name=metadata_file)
            except XLRDError:
                element_description = pd.read_excel(element_description_file, sheet_name=metadata_file,
                                                    engine='openpyxl')

            print("metadata_file: " + str(metadata_file))
            for index, row in element_description.iterrows():
                print(str(row["Element"]))
                print("    Required: " + str(row["Required"]))
                print("    Type: " + str(row["Type"]))
                print("    Description: " + str(row["Description"]))
                print("    Example: " + str(row["Example"]))

            fields = element_description.values.tolist()
            return fields

        if not self._template:
            self._load_template(version=None)

        data = self._template.get(metadata_file)
        metadata = data.get("metadata")
        # set the first column as the index column
        metadata = metadata.set_index(list(metadata)[0])
        if axis == 0:
            fields = list(metadata.columns)
        elif axis == 1:
            fields = list(metadata.index)

        print("Fields:")
        for field in fields:
            print(field)

        return fields

    def _generate_metadata(self):
        metadata_files = self.list_metadata_files(self._version, print_list=False)
        for metadata_file in metadata_files:
            metadata = self._dataset.get(metadata_file).get("metadata")
            self._metadata[metadata_file] = Metadata(metadata_file, metadata, self._version, self._dataset_path)
            if metadata_file == "subjects":
                Subject._metadata = self._metadata[metadata_file]
            elif metadata_file == 'samples':
                Sample._metadata = self._metadata[metadata_file]

        Sample._manifest_metadata = self._metadata['manifest']

    def get_metadata(self, metadata_file):
        """
        Get a Metadata object based on the metadata file name
        To edit values for a metadata

        :param metadata_file: one of string of [code_description,
                                                code_parameters,
                                                dataset_description,
                                                manifest,performances,
                                                resources,samples,
                                                subjects,submission]
        :type  metadata_file: string
        :return: give a metadata editor for a specific metadata
        """
        if not self._dataset:
            msg = "Dataset not defined. Please load the dataset in advance."
            raise ValueError(msg)

        metadata_file = validate_metadata_file(metadata_file, self._version)
        return self._metadata[metadata_file]

    def _set_field(self, metadata_file, row_index, header, value):
        """
        Set single field by row idx/name and column name (the header)
        TODO: will delete later

        :param metadata_file: metadata metadata_file
        :type metadata_file: string
        :param row_index: row index in Excel. Excel index starts from 1 where index 1 is the header row. so actual data index starts from 2
        :type row_index: int
        :param header: column name. the header is the first row
        :type header: string
        :param value: field value
        :type value: string
        :return: updated dataset
        :rtype: dict
        """
        if not self._dataset:
            msg = "Dataset not defined. Please load the dataset in advance."
            raise ValueError(msg)

        metadata = self._dataset.get(metadata_file).get("metadata")

        if not isinstance(row_index, int):
            msg = "row_index should be 'int'."
            raise ValueError(msg)

        try:
            # Convert Excel row index to dataframe index: index - 2
            row_index = row_index - 2
            metadata.loc[row_index, header] = value
        except ValueError:
            msg = "Value error. row does not exists."
            raise ValueError(msg)

        self._dataset[metadata_file]["metadata"] = metadata

        return self._dataset

    def _set_field_using_row_name(self, metadata_file, row_name, header, value):
        """
        Set single cell. The row is identified by the given unique name and column is identified by the header.

        :param metadata_file: metadata metadata_file
        :type metadata_file: string
        :param row_name: Unique row name in Excel. (Ex: if subjects is metadata_file, a row name can be a unique subjet id)
        :type row_name: string
        :param header: column name. the header is the first row
        :type header: string
        :param value: field value
        :type value: string
        :return: updated dataset
        :rtype: dict
        """
        if not self._dataset:
            msg = "Dataset not defined. Please load the dataset in advance."
            raise ValueError(msg)

        metadata = self._dataset.get(metadata_file).get("metadata")

        if not isinstance(row_name, str):
            msg = "row_name should be string."
            raise ValueError(msg)

        # Assumes that all excel files first column contains the unique value field
        # TODO: In version 1, the unique column is not the column 0. Hence, unique column must be specified. 
        # This method need to be fixed to accomadate this 
        matching_indices = metadata.index[metadata[metadata.columns[0]] == row_name].tolist()

        if not matching_indices:
            msg = f"No row with given unique name, {row_name}, was found in the unique column {metadata.columns[0]}"
            raise ValueError(msg)
        elif len(matching_indices) > 1:
            msg = f"More than one row with given unique name, {row_name}, was found in the unique column {metadata.columns[0]}"
            raise ValueError(msg)
        else:
            excel_row_index = matching_indices[0] + 2
            return self._set_field(metadata_file=metadata_file, row_index=excel_row_index, header=header, value=value)

    def _append(self, metadata_file, row, check_exist=False, unique_column=None):
        """
        Append a row to a metadata file

        :param metadata_file: metadata metadata_file
        :type metadata_file: string
        :param row: a row to be appended
        :type row: dic
        :param check_exist: Check if row exist before appending, if exist, update row, defaults to False
        :type check_exist: bool, optional
        :param unique_column: if check_exist is True, provide which column in metadata_file is unique, defaults to None
        :type unique_column: string, optional
        :raises ValueError: _description_
        :return: updated dataset
        :rtype: dict
        """
        if not self._dataset:
            msg = "Dataset not defined. Please load the dataset in advance."
            raise ValueError(msg)

        # metadata = self._dataset.get(metadata_file).get("metadata")
        current_metadata = self.get_metadata(metadata_file)
        if check_exist:
            # In version 1, the unique column is not the column 0. Hence, unique column must be specified
            if unique_column is None:
                error_msg = "Provide which column in metadata_file is unique. Ex: subject_id"
                raise ValueError(error_msg)

            try:
                row_index = check_row_exist(current_metadata.data, unique_column, unique_value=row[unique_column])
            except ValueError:
                error_msg = "Row values provided does not contain a unique identifier"
                raise ValueError(error_msg)
        else:
            row_index = -1

        if row_index == -1:
            # Add row
            row_df = pd.DataFrame([row])
            current_metadata.data = pd.concat([current_metadata.data, row_df], axis=0,
                                              ignore_index=True)  # If new header comes, it will be added as a new column with its value
        else:
            # Append row with additional values
            for key, value in row.items():
                current_metadata.data.loc[row_index, key] = value

        self._dataset[metadata_file]["metadata"] = current_metadata.data
        return self._dataset

    def update_by_json(self, metadata_file, json_file):
        """
        Given json file, update metadata file
        :param metadata_file: metadata metadata_file/filename
        :type metadata_file: string
        :param json_file: path to metadata file in json
        :type json_file: string
        :return:
        :rtype:
        """
        metadata = self._dataset.get(metadata_file).get("metadata")

        with open(json_file, "r") as f:
            data = json.load(f)

        for key, value in data.items():
            if isinstance(value, dict):
                for key_1, value_1 in value.items():
                    if isinstance(value, list):
                        field = "    " + key_1
                        value = str(value_1)
                    else:
                        field = "    " + key_1
                        value = value_1

                    index = metadata.index[metadata['Metadata element'] == field].tolist()[0]
                    metadata.loc[index, "Value"] = value

            elif isinstance(value, list):
                field = key
                value = str(value)
                index = metadata.index[metadata['Metadata element'] == field].tolist()[0]
                metadata.loc[index, "Value"] = value
            else:
                field = key
                index = metadata.index[metadata['Metadata element'] == field].tolist()[0]
                metadata.loc[index, "Value"] = value

        return metadata

    def _generate_file_from_template(self, save_path, metadata_file, data=pd.DataFrame(), keep_style=False):
        """Generate file from a template and populate with data if givn
        TODO: will delete later

        :param save_path: destination to save the generated file
        :type save_path: string
        :param metadata_file: SDS metadata_file (Ex: samples, subjects)
        :type metadata_file: string
        :param data: pandas dataframe containing data, defaults to pd.DataFrame()
        :type data: pd.DataFrame, optional
        """

        if keep_style:
            self._template_dir = self._get_template_dir(version=self._version)
            sf = StyleFrame.read_excel_as_template(os.path.join(self._template_dir, f'{metadata_file}.xlsx'), data)
            writer = StyleFrame.ExcelWriter(save_path)
            sf.to_excel(writer)
            writer.save()
        else:
            data.to_excel(save_path, index=False)

    """***************************New Add subjects ***************************"""

    def add_subjects(self, subjects):

        """
        Add Subejct list to dataset.
        This function will add subjects and samples to metadata,
        And will move the sample files from origin source path to dataset
        primary subject sample folder.
        It will automatically update manifest and dataset_description metadata files.

        :param subjects: Subject dataset
        :type subjects: list
        """

        self.save()
        if not isinstance(subjects, list):
            msg = "Please provide a list of subjects"
            raise ValueError(msg)
        for subject in subjects:
            self._subjects[subject.subject_id] = subject
            subject.move()

        self._update_sub_sam_nums_in_dataset_description(self._dataset_path / 'primary')

    def get_subject(self, subject_sds_id) -> Subject:
        """
        Get a subject by subject sds id

        :param subject_sds_id: subject sds id
        :type subject_sds_id: str
        :return: Subject
        """
        if not isinstance(subject_sds_id, str):
            msg = f"Subject not found, please provide a string subject_sds_id!, you subject_sds_id type is {type(subject_sds_id)}"
            raise ValueError(msg)

        try:
            subject = self._subjects.get(subject_sds_id)
            return subject
        except:
            msg = f"Subject not found with {subject_sds_id}! Please check your subject_sds_id in subject metadata file"
            raise ValueError(msg)

    def add_derivative_data(self, source_path, subject, sample, copy=True, overwrite=True):
        """Add raw data of a sample to correct SDS location and update relavent metadata files.
        Requires you to already have the folder structure inplace.

        :param source_path: original location of raw data
        :type source_path: string
        :param subject: subject id
        :type subject: string
        :param sample: sample id
        :type sample: string
        :param sds_parent_dir: path to existing sds dataset parent
        :type sds_parent_dir: string, optional
        :param copy: if True, source directory data will not be deleted after copying, defaults to True
        :type copy: bool, optional
        :param overwrite: if True, any data in the destination folder will be overwritten, defaults to False
        :type overwrite: bool, optional
        :raises NotADirectoryError: if the derivative in sds_parent_dir is not a folder, this wil be raised.
        """

        derivative_folder = os.path.join(str(self._dataset_path), 'derivative')

        # Check if sds_parent_directory contains the derivative folder. If not create it.
        if os.path.exists(derivative_folder):
            if not os.path.isdir(derivative_folder):
                raise NotADirectoryError(f'{derivative_folder} is not a directory')
        else:
            os.mkdir(derivative_folder)

        self._add_sample_data(source_path, self._dataset_path, subject, sample, data_type="derivative", copy=copy,
                              overwrite=overwrite)

    def _add_element(self, metadata_file, element):
        """
        May need to delete

        :param metadata_file:
        :param element:
        :return:
        """
        metadata = self._dataset.get(metadata_file).get("metadata")
        if metadata_file in self._column_based:
            row_pd = pd.DataFrame([{"Metadata element": element}])
            metadata = pd.concat([metadata, row_pd], axis=0, ignore_index=True)
        else:
            metadata[element] = None

        self._dataset[metadata_file]["metadata"] = metadata

    def add_thumbnail(self, source_path, copy=True, overwrite=True):

        file_source_path = Path(source_path)
        if not file_source_path.is_file():
            msg = f"source_path should be the thumbnail file's path"
            raise ValueError(msg)
        else:
            filename = file_source_path.name
            destination_path = self._dataset_path.joinpath('docs', filename)
            if destination_path.exists():
                if overwrite:
                    self._delete_data(destination_path)
                else:
                    msg = f"The thumbnail file has already in primary folder"
                    raise FileExistsError(msg)

            self._move_single_file(file_path=source_path, destination_path=destination_path, fname=filename, copy=copy)
            description = f"This is a thumbnail file"
            self._modify_manifest(fname=filename, manifest_folder_path=str(self._dataset_path),
                                  destination_path=str(destination_path.parent), description=description)


    def _add_sample_data(self, source_path, dataset_path, subject, sample, data_type="primary", copy=True,
                         overwrite=True):
        """Copy or move data from source folder to destination folder

        :param source_path: path to the original data
        :type source_path: string
        :param destination_path_list: folder path in a list[root, data_pype, subject, sample] to be copied into
        :type destination_path_list: list
        :param copy: if True, source directory data will not be deleted after copying, defaults to True
        :type copy: bool, optional
        :param overwrite: if True, any data in the destination folder will be overwritten, defaults to False
        :type overwrite: bool, optional
        :raises FileExistsError: if the destination folder contains data and overwritten is set to False, this wil be raised.
        """
        destination_path = os.path.join(str(dataset_path), data_type, subject, sample)
        # If overwrite is True, remove existing sample
        if os.path.exists(destination_path):
            if os.path.isdir(source_path):
                if overwrite:
                    shutil.rmtree(destination_path)
                    os.makedirs(destination_path)
                else:
                    raise FileExistsError(
                        "Destination file already exist. Indicate overwrite argument as 'True' to overwrite the existing")
            else:
                if overwrite:
                    file_path = Path(destination_path).joinpath(Path(source_path).name)
                    self._delete_data(file_path)
                else:
                    raise FileExistsError(
                        "Destination file already exist. Indicate overwrite argument as 'True' to overwrite the existing")
        else:
            # Create destination folder
            os.makedirs(destination_path)

        description = f"File of subject {subject} sample {sample}"
        if os.path.isdir(source_path):
            for fname in os.listdir(source_path):
                file_path = os.path.join(source_path, fname)
                if os.path.isdir(file_path):
                    # Warn user if a subdirectory exist in the input_path
                    print(
                        f"Warning: Input directory consist of subdirectory {source_path}. It will be avoided during copying")
                    return
                else:
                    self._move_single_file(file_path=file_path, destination_path=destination_path,
                                           fname=fname, copy=copy)
                    self._modify_manifest(fname=fname, manifest_folder_path=dataset_path,
                                          destination_path=destination_path,
                                          description=description)
        else:
            fname = os.path.basename(source_path)
            self._move_single_file(file_path=source_path, destination_path=destination_path,
                                   fname=fname, copy=copy)
            self._modify_manifest(fname=fname, manifest_folder_path=dataset_path, destination_path=destination_path,
                                  description=description)

    def _move_single_file(self, file_path, destination_path, fname, copy):
        if copy:
            # Copy data
            shutil.copy2(file_path, destination_path)
        else:
            # Move data
            shutil.move(file_path, os.path.join(destination_path, fname))

    def _modify_manifest(self, fname, manifest_folder_path, destination_path, description=""):
        # Check if manifest exist
        # If can be "xlsx", "csv" or "json"
        files = os.listdir(manifest_folder_path)
        manifest_file_path = [f for f in files if "manifest" in f]
        # Case 1: manifest file exists
        if len(manifest_file_path) != 0:
            manifest_file_path = os.path.join(manifest_folder_path, manifest_file_path[0])
            # Check the extension and read file accordingly
            extension = os.path.splitext(manifest_file_path)[-1].lower()
            if extension == ".xlsx":
                df = pd.read_excel(manifest_file_path)
            elif extension == ".csv":
                df = pd.read_csv(manifest_file_path)
            elif extension == ".json":
                # TODO: Check what structure a manifest json is in
                # Below code assumes json structure is like
                # '{"row 1":{"col 1":"a","col 2":"b"},"row 2":{"col 1":"c","col 2":"d"}}'
                df = pd.read_json(manifest_file_path, orient="index")
            else:
                raise ValueError(f"Unauthorized manifest file extension: {extension}")
        # Case 2: create manifest file
        else:
            # Default extension to xlsx
            extension = ".xlsx"
            # Creat manifest file path
            manifest_file_path = os.path.join(manifest_folder_path, "manifest.xlsx")
            df = pd.DataFrame(columns=['filename', 'description', 'timestamp', 'file type'])

        file_path = Path(
            str(os.path.join(destination_path, fname)).replace(str(manifest_folder_path), '')[1:]).as_posix()

        row = {
            'filename': file_path,
            'timestamp': datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
            'description': description,
            'file type': os.path.splitext(fname)[-1].lower()[1:]
        }

        exsiting_row = df['filename'] == row['filename']
        if exsiting_row.any():
            df.loc[exsiting_row, 'timestamp'] = row['timestamp']
        else:
            row_pd = pd.DataFrame([row])
            df = pd.concat([df, row_pd], axis=0, ignore_index=True)

        # update dataset metadata
        self._update_dataset_by_df(df, "manifest")

        # Save editted manifest file
        if extension == ".xlsx":
            df.to_excel(manifest_file_path, index=False)
        elif extension == ".csv":
            df = pd.to_csv(manifest_file_path, index=False)
        elif extension == ".json":
            df = pd.read_json(manifest_file_path, orient="index")
        return

    def _update_dataset_by_df(self, df, metadata_file):
        manifest_metadata = self._metadata[metadata_file]
        manifest_metadata.data = df
        self._dataset[metadata_file]["metadata"] = manifest_metadata.data

    """************************************ Delete Data Functions ************************************"""

    def delete_subjects(self, destination_paths, data_type="primary"):
        """
        :param destination_paths: the subject folder paths that you want to delete
        :type destination_paths: str[]
        :param data_type: "primary" | "derivative"
        :type: str
        :return:
        """
        if isinstance(destination_paths, list):
            for sub_folder in destination_paths:
                self.delete_subject(destination_path=sub_folder, data_type=data_type)
        else:
            msg = f"Please provide a list, and put all your deleting sample paths in a list"
            raise ValueError(msg)

    def delete_subject(self, destination_path, data_type="primary"):
        """
        :param destination_path: the subject folder path that you want to delete
        :type destination_path: str
        :param data_type: "primary" | "derivative"
        :type: str
        :return:
        """
        if isinstance(destination_path, list):
            msg = f"Please provide a path string!"
            raise ValueError(msg)

        sub_folder = Path(destination_path)
        if not sub_folder.exists():
            msg = f"The folder {sub_folder} is not existing"
            raise ValueError(msg)
        elif not sub_folder.is_dir():
            msg = f"The {sub_folder} is not a folder"
            raise ValueError(msg)
        else:
            primary_folder = self._dataset_path / "primary"
            for sam_folder in sub_folder.iterdir():
                if sam_folder.is_dir():
                    self.delete_sample(sam_folder, data_type)

            sub_folder.rmdir()
            if data_type == "primary":
                self._update_sub_sam_nums_in_dataset_description(primary_folder)
                subjects_metadata = self._metadata["subjects"]
                subjects_metadata.remove_row(sub_folder.name)
                subjects_metadata.save()

    def delete_samples(self, destination_paths, data_type="primary"):
        """
        :param destination_paths: a list of deleting sample folders
        :type destination_paths: list
        :param data_type: "primary" | "derivative"
        :type data_type: str
        :return:
        """
        if isinstance(destination_paths, list):
            for sam_folder in destination_paths:
                self.delete_sample(destination_path=sam_folder, data_type=data_type)
        else:
            msg = f"The {destination_paths} type is {type(destination_paths)}. Please provide a list, and put all your deleting sample paths in a list"
            raise TypeError(msg)

    def delete_sample(self, destination_path, data_type="primary"):
        """
        :param destination_path: the sample folder path that you want to delete
        :param data_type:
        :return:
        """
        if isinstance(destination_path, list):
            msg = f"Please provide a path string!"
            raise TypeError(msg)

        sam_folder = Path(destination_path)
        if not sam_folder.exists():
            msg = f"The folder {sam_folder} is not existing"
            raise FileExistsError(msg)
        elif not sam_folder.is_dir():
            msg = f"The {sam_folder} path is not a folder, please provide the sample files folder."
            raise ValueError(msg)
        else:
            primary_folder = self._dataset_path / "primary"
            for item in sam_folder.iterdir():
                self.delete_data(item)
            sam_folder.rmdir()
            if data_type == "primary":
                self._update_sub_sam_nums_in_dataset_description(primary_folder)
                samples_metadata = self._metadata["samples"]
                samples_metadata.remove_row(sam_folder.name)
                samples_metadata.save()

    def remove_thumbnail(self, destination_path):
        """
        Delete a thumbnail from dataset
        Will automatically update manifest metadata.

        :param destination_path: The thumbnail path in dataset that you want to delete.
        :type destination_path: str
        """
        self.delete_data(destination_path)

    def delete_data(self, destination_path):
        """
        Delete file based on ,the file path in dataset
        It will automatically update mainfest metadata
        TODO: need to connect delete sample and subject, and update subject and sample metadata

        :param destination_path: the file path that you want to delete
        :type destination_path: str
        :return:
        """
        if not Path(destination_path).exists():
            msg = f"The file {str(destination_path)} is not existing"
            raise FileNotFoundError(msg)
        else:
            delete_flag = self._delete_data(destination_path)
            if delete_flag:
                path = str(Path(str(Path(destination_path).as_posix()).replace(str(self._dataset_path.as_posix()), "")[
                                1:]).as_posix())
                manifest = self._metadata["manifest"]
                manifest.remove_row(path)
                manifest.save()

    def _delete_data(self, destination_path):
        file_path = Path(destination_path)
        if file_path.exists():
            if file_path.is_file():
                file_path.unlink()
                return True
            else:
                shutil.rmtree(file_path)
                return False
        else:
            return False

    def _update_sub_sam_nums_in_dataset_description(self, primary_folder):
        """
        :param primary_folder: the primary folder url
        :type: Path|str
        :return:
        """
        subject_folders = get_sub_folder_paths_in_folder(primary_folder)
        sample_folders = []
        for sub in subject_folders:
            if sub.is_dir():
                folders = get_sub_folder_paths_in_folder(sub)
                sample_folders.extend(folders)
        dataset_description_metadata = self._metadata["dataset_description"]
        dataset_description_metadata.set_values(element="Number of subjects", values=len(subject_folders))
        dataset_description_metadata.set_values(element="Number of samples", values=len(sample_folders))
        dataset_description_metadata.save()



import json
import requests

# import gen3
from gen3.auth import Gen3Auth
from gen3.submission import Gen3Submission


def graphql_api():
    # project_id = "demo1-12L"
    #
    # token = get_api_token(cred_file=cred_file, url=BASE_URL)
    # headers = {'Authorization': 'bearer ' + token['access_token']}
    # print("Headers:" + str(headers))
    #
    # # get projects
    # query = {'query': """{project(first:0){project_id id}}"""}
    # url = BASE_URL + "/api/v0/submission/graphql/"
    #
    # ql = requests.post(url, json=query, headers=headers)
    # print(ql.text)
    #
    # print("done")

    # "http://gen3.abi-ctt-ctp.cloud.edu.au/data/download/ca59f43f-362b-4179-b71e-963a7dcc22ba"
    pass


def get_api_token(cred_file, url):
    with open(cred_file) as f:
        credentials = json.load(f)
    key = {
        "api_key": credentials.get("api_key"),
        "key_id": credentials.get("key_id")
    }
    token = requests.post(url + "/user/credentials/cdis/access_token", json=key).json()

    return token


def init_submisstor(cred_file, base_url):
    auth = Gen3Auth(base_url, refresh_file=cred_file)
    sub = Gen3Submission(auth)

    return sub


def query_graphql(sub, query_string, variables=None):
    responses = sub.query(query_string, variables).get("data")

    return responses


def get_experiemnts(sub, experiments=None):
    pass

def get_subjects(sub, experiments=None, subjects=None):
    pass

def get_samples(sub, experiments=None, subjects=None, samples=None):
    pass

def get_programs(sub):
    response = sub.get_programs()

    programs = list()
    for resp in response.get("links"):
        program = resp.split('/')[-1]
        programs.append(program)

    return programs


def get_projects(sub, program):
    response = sub.get_projects(program)

    projects = list()
    for resp in response.get("links"):
        project = resp.split('/')[-1]
        projects.append(project)

    return projects


def get_node_records(sub, node, program, project):
    resp = sub.export_node(program, project, node, "json")
    records = resp.get("data")
    return records


if __name__ == '__main__':
    cred_file = "credentials.json"
    BASE_URL = "https://gen3.abi-ctt-ctp.cloud.edu.au"

    sub = init_submisstor(cred_file, BASE_URL)

    programs = get_programs(sub)
    print("Programs: " + str(programs))
    projects = get_projects(sub, programs[0])
    print(str(programs[0]) + ": " + str(projects))

    experiments = get_node_records(sub, "experiment", programs[0], projects[0])
    print(experiments)

    subjects = get_node_records(sub, "case", programs[0], projects[0])
    print(subjects)

    print("GraphQL")
    query_string = """
        { project(first:0) { code } }
    """
    query_string = """
        { project{ name } }
    """
    query_string = """
        { experiment{ submitter_id }}
    """
    query_string = """
        { 
            experiment (submitter_id: "dataset_duke"){
                submitter_id
            }
        }
    """
    query_string = """
        {
          experiment(submitter_id: "1.3.6.1.4.1.14519.5.2.1.186051521067863971269584893740842397538"){
            id
            submitter_id
            cases{
              id
              submitter_id,
              subject_id,
              samples{
                id
                submitter_id,
                sample_id
              }
            }
          }
        }
    """

    query_string = """
            {
              experiment(submitter_id: "1.3.6.1.4.1.14519.5.2.1.186051521067863971269584893740842397538"){
                id,
                submitter_id
                manifests{
                  id,
                  submitter_id
                }
              }
            }
        """
    results = query_graphql(sub, query_string)
    print(results)



"""
Example for generating a SDS derived dataset dataset and filling the dataset_description
"""

from sparc_me import Dataset
from sparc_me.core.utils import add_data

if __name__ == '__main__':
    # dataset_path = "./tmp/example/derived_dataset/"
    # file_path = "./tmp/example/derived_dataset/tmp.txt"

    dataset_path = "C:\\Users\\clin864\\ctt_group\\sparc_datasets\\dataset-12L_1-version-1 - Copy"
    file_path = "C:\\Users\\clin864\\ctt_group\\sparc_datasets\\tmp.txt"

    dataset = Dataset()

    # Creating a SDS dataset
    dataset.load_from_template(version="2.0.0")
    dataset.save(save_dir=dataset_path)

    # load dataset
    # dataset1 = dataset.load_dataset(dataset_path)

    # Filling/Updating dataset.

    # Filling dataset description
    # the values can be filled in by 2 methods, set_field() or set_field_using_row_name().
    # Using Dataset.set_field()
    # You can get the row_index by looking at
    #   1. the saved metadata file dataset_description.xlsx. Excel index starts from 1 where index 1 is the header row. so actual data index starts from 2.
    #   2. or the DataFrame object in the python code. dataset._dataset.dataset_description.metadata
    #
    dataset.set_field(category="dataset_description", row_index=2, header="Value", value="2.0.0")
    dataset.set_field(category="dataset_description", row_index=3, header="Value", value="experimental")
    dataset.set_field(category="dataset_description", row_index=5, header="Value", value="Duke Breast Cancer MRI segmentation")
    dataset.set_field(category="dataset_description", row_index=6, header="Value", value="NA")
    dataset.set_field(category="dataset_description", row_index=7, header="Value", value="Breast cancer")
    dataset.set_field(category="dataset_description", row_index=11, header="Value", value="Segmenting breast MRI")
    dataset.set_field(category="dataset_description", row_index=12, header="Value", value="derived from Duke Breast Cancer MRI dataset")
    dataset.set_field(category="dataset_description", row_index=13, header="Value", value="NA")
    dataset.set_field(category="dataset_description", row_index=14, header="Value", value="breast")
    dataset.set_field(category="dataset_description", row_index=15, header="Value", value="Machine-learning")
    dataset.set_field(category="dataset_description", row_index=16, header="Value", value="Tensorflow 1")
    dataset.set_field(category="dataset_description", row_index=17, header="Value", value="NA")
    dataset.set_field(category="dataset_description", row_index=19, header="Value", value="Lin, Chinchien")
    dataset.set_field(category="dataset_description", row_index=20, header="Value", value="https://orcid.org/0000-0001-8170-199X")
    dataset.set_field(category="dataset_description", row_index=21, header="Value", value="University of Auckland")
    dataset.set_field(category="dataset_description", row_index=22, header="Value", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=24, header="Value", value="source")
    dataset.set_field(category="dataset_description", row_index=25, header="Value", value="WasDerivedFrom")
    dataset.set_field(category="dataset_description", row_index=26, header="Value", value="https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70226903")
    dataset.set_field(category="dataset_description", row_index=27, header="Value", value="URL")
    dataset.set_field(category="dataset_description", row_index=29, header="Value", value="923")
    dataset.set_field(category="dataset_description", row_index=30, header="Value", value="5032")

    # Saving the updated dataset
    dataset.save(dataset_path)

    # adding data
    add_data(file_path, dataset_path, copy=True, overwrite=False)



from pathlib import Path
import yaml
if __name__ == '__main__':
    path = Path(r"C:\Users\clin864\Desktop\sparc-codathon\sparc-me\my-scripts\gen3\sds_dictionary\case.yaml")
    with open(path, 'r') as stream:
        schema = yaml.safe_load(stream)

    pass



from pathlib import Path
import json
import os

if __name__ == '__main__':
    experiment = "dataset-264-version-1"
    dest = Path(r"C:\Users\clin864\ctt_group\sparc_datasets")
    dest = dest.joinpath(dest, experiment)
    data = {
        "type": "experiment",
        "submitter_id": experiment,
        "projects": [
            {
                "code": "12L"
            }
        ]
    }

    # saving
    os.makedirs(dest, exist_ok=True)

    dest = dest.joinpath("experiment.json")
    with open(dest, 'w') as f:
        json.dump(data, f, indent=4)
    print("Saved to " + str(dest))



"""
Example for generating a SDS primary dataset dataset and filling the dataset_description
"""

from sparc_me import Dataset

if __name__ == '__main__':
    save_dir = "./tmp/example/primary_dataset/"

    # Creating a SDS dataset
    dataset = Dataset()
    dataset.load_from_template(version="2.0.0")
    dataset.save(save_dir=save_dir)


    # Filling/Updating dataset.

    # Filling dataset description
    # the values can be filled in by 2 methods, set_field() or set_field_using_row_name().
    # Using Dataset.set_field()
    # You can get the row_index by looking at
    #   1. the saved metadata file dataset_description.xlsx. Excel index starts from 1 where index 1 is the header row. so actual data index starts from 2.
    #   2. or the DataFrame object in the python code. dataset._dataset.dataset_description.metadata
    #
    dataset.set_field(category="dataset_description", row_index=2, header="Value", value="2.0.0")
    dataset.set_field(category="dataset_description", row_index=3, header="Value", value="experimental")
    dataset.set_field(category="dataset_description", row_index=5, header="Value", value="Dynamic contrast-enhanced magnetic resonance images of breast cancer patients with tumor locations (Duke-Breast-Cancer-MRI)")
    dataset.set_field(category="dataset_description", row_index=6, header="Value", value="""Background Recent studies showed preliminary data on associations of MRI-based imaging phenotypes of breast tumours with breast cancer molecular, genomic, and related characteristics. In this study, we present a comprehensive analysis of this relationship.
Methods We analysed a set of 922 patients with invasive breast cancer and pre-operative MRI. The MRIs were analysed by a computer algorithm to extract 529 features of the tumour and the surrounding tissue. Machine-learning-based models based on the imaging features were trained using a portion of the data (461 patients) to predict the following molecular, genomic, and proliferation characteristics: tumour surrogate molecular subtype, oestrogen receptor, progesterone receptor and human epidermal growth factor status, as well as a tumour proliferation marker (Ki-67). Trained models were evaluated on the set of the remaining 461 patients.
Results Multivariate models were predictive of Luminal A subtype with AUC = 0.697 (95% CI: 0.647–0.746, p < .0001), triple negative breast cancer with AUC = 0.654 (95% CI: 0.589–0.727, p < .0001), ER status with AUC = 0.649 (95% CI: 0.591–0.705, p < .001), and PR status with AUC = 0.622 (95% CI: 0.569–0.674, p < .0001). Associations between individual features and subtypes we also found.
Conclusions There is a moderate association between tumour molecular biomarkers and algorithmically assessed imaging features.""")
    dataset.set_field(category="dataset_description", row_index=7, header="Value", value="Breast cancer")
    dataset.set_field(category="dataset_description", row_index=7, header="Value 2", value="MRI")
    dataset.set_field(category="dataset_description", row_index=11, header="Value", value="""Breast MRI is a common image modality to assess the extent of disease in breast cancer patients. Recent studies show that MRI has a potential in prognosis of patients’ short and long-term outcomes as well as predicting pathological and genomic features of the tumors. However, large, well annotated datasets are needed to make further progress in the field. We share such a dataset here.""")
    dataset.set_field(category="dataset_description", row_index=12, header="Value", value="""In terms of design, the dataset is a single-institutional, retrospective collection of 922 biopsy-confirmed invasive breast cancer patients, over a decade, having the following data components:
Demographic, clinical, pathology, treatment, outcomes, and genomic data: Collected from a variety of sources including clinical notes, radiology report, and pathology reports and has served as a source for multiple published papers on radiogenomics, outcomes prediction, and other areas.
Pre-operative dynamic contrast enhanced (DCE)-MRI: Downloaded from PACS systems and de-identified for The Cancer Imaging Archive (TCIA) release. These include axial breast MRI images acquired by 1.5T or 3T scanners in the prone positions. Following MRI sequences are shared in DICOM format: a non-fat saturated T1-weighted sequence, a fat-saturated gradient echo T1-weighted pre-contrast sequence, and mostly three to four post-contrast sequences.
Locations of lesions in DCE-MRI: Annotations on the DCE-MRI images by radiologists.
Imaging features from DCE-MRI: A set of 529 computer-extracted imaging features by inhouse software. These features represent a variety of imaging characteristics including size, shape, texture, and enhancement of both the tumor and the surrounding tissue, which is combined of features commonly published in the literature, as well as the features developed in our lab.""")
    dataset.set_field(category="dataset_description", row_index=13, header="Value", value="There is a moderate association between tumour molecular biomarkers and algorithmically assessed imaging features.")
    dataset.set_field(category="dataset_description", row_index=14, header="Value", value="breast")
    dataset.set_field(category="dataset_description", row_index=15, header="Value", value="Machine-learning")
    dataset.set_field(category="dataset_description", row_index=16, header="Value", value="""We analysed a set of 922 patients with invasive breast cancer and pre-operative MRI. The MRIs were analysed by a computer algorithm to extract 529 features of the tumour and the surrounding tissue. Machine-learning-based models based on the imaging features were trained using a portion of the data (461 patients) to predict the following molecular, genomic, and proliferation characteristics: tumour surrogate molecular subtype, oestrogen receptor, progesterone receptor and human epidermal growth factor status, as well as a tumour proliferation marker (Ki-67). Trained models were evaluated on the set of the remaining 461 patients.""")
    dataset.set_field(category="dataset_description", row_index=17, header="Value", value="Dynamic contrast-enhanced magnetic resonance images of breast cancer patients with tumor locations (Duke-Breast-Cancer-MRI)")
    dataset.set_field(category="dataset_description", row_index=19, header="Value", value="Saha, Ashirbani")
    dataset.set_field(category="dataset_description", row_index=19, header="Value 2", value="Harowicz, Michael R")
    dataset.set_field(category="dataset_description", row_index=19, header="Value 3", value="Grimm, Lars J")
    dataset.set_field(category="dataset_description", row_index=19, header="Value 4", value="Kim, Connie E")
    dataset.set_field(category="dataset_description", row_index=19, header="Value 5", value="Ghate, Sujata V")
    dataset.set_field(category="dataset_description", row_index=19, header="Value 6", value="Walsh, Ruth")
    dataset.set_field(category="dataset_description", row_index=19, header="Value 7", value="Mazurowski, Maciej A")
    dataset.set_field(category="dataset_description", row_index=20, header="Value", value="https://orcid.org/0000-0002-7650-1720")
    dataset.set_field(category="dataset_description", row_index=20, header="Value 2", value="https://orcid.org/0000-0002-8002-5210")
    dataset.set_field(category="dataset_description", row_index=20, header="Value 3", value="https://orcid.org/0000-0002-3865-3352")
    dataset.set_field(category="dataset_description", row_index=20, header="Value 4", value="https://orcid.org/0000-0003-0730-0551")
    dataset.set_field(category="dataset_description", row_index=20, header="Value 5", value="https://orcid.org/0000-0003-1889-982X")
    dataset.set_field(category="dataset_description", row_index=20, header="Value 6", value="https://orcid.org/0000-0002-2164-2761")
    dataset.set_field(category="dataset_description", row_index=20, header="Value 7", value="https://orcid.org/0000-0003-4202-8602")
    dataset.set_field(category="dataset_description", row_index=21, header="Value", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=21, header="Value 2", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=21, header="Value 3", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=21, header="Value 4", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=21, header="Value 5", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=21, header="Value 6", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=21, header="Value 7", value="Duke University")
    dataset.set_field(category="dataset_description", row_index=22, header="Value", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=22, header="Value 2", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=22, header="Value 3", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=22, header="Value 4", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=22, header="Value 5", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=22, header="Value 6", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=22, header="Value 7", value="Researcher")
    dataset.set_field(category="dataset_description", row_index=24, header="Value", value="source")
    dataset.set_field(category="dataset_description", row_index=25, header="Value", value="IsDescribedBy")
    dataset.set_field(category="dataset_description", row_index=26, header="Value", value="9d70fd9f-bfb9-424d-9c7c-9db1ec6a9df9")
    dataset.set_field(category="dataset_description", row_index=27, header="Value", value="12L digital twin UUID")
    dataset.set_field(category="dataset_description", row_index=29, header="Value", value="1")
    dataset.set_field(category="dataset_description", row_index=30, header="Value", value="1")

    # Saving the updated dataset
    dataset.save(save_dir)



import json
import pandas as pd
from pathlib import Path
from typing import Dict
import openpyxl
import os
import math

from sparc_me.core.utils import CaseInsensitiveDict, validate_metadata_file

from xlrd import XLRDError

from jsonschema import validate
from jsonschema.exceptions import ValidationError
from jsonschema import Draft202012Validator
from jsonschema.exceptions import best_match

current_dir = Path(__file__).parent.resolve()
resources_dir = Path.joinpath(current_dir, "../resources")


def convert_version_format(version):
    """
    Convert version format
    :param version: dataset/template version
    :type version: string
    :return: version in the converted format
    :rtype:
    """
    version = version.replace(".", "_")

    if "_" not in version:
        version = version + "_0_0"

    return version


class Validator(object):
    def __init__(self):
        self._version = None
        self._metadata_file = None
        self._schema_ref = None

    def validate_dataset(self, dataset):

        metadata_files = dataset.list_metadata_files(version="2.0.0", print_list=False)
        currently_support_metadata_files = ['dataset_description', 'manifest', 'samples', 'subjects']
        schema = Schema()

        for metadata_file in currently_support_metadata_files:
            metadata = dataset.get_metadata(metadata_file)
            data = schema.load_data(metadata.metadata_file_path)
            self.validate(data, metadata_file=metadata_file, version=metadata.version)

    def validate(self, data, metadata_file, version):
        """
        Validate data instance
        :param data: Target data
        :type data: dict or list
        :param metadata_file: metadata filename
        :type metadata_file: string
        :param version: dataset version
        :type version: string
        :return:
        :rtype:
        """
        self._metadata_file = metadata_file
        self._version = convert_version_format(version)
        self._load_reference_schema()
        if isinstance(data, dict):
            self._execute(data)
        elif isinstance(data, list):
            for idx in range(len(data)):
                self._execute(data[idx])
        else:
            print("Input data type invalid")

    def _load_reference_schema(self):
        """
        Load the reference schema which will be used to validate the target metadata
        :return:
        :rtype:
        """
        version = "version_" + self._version
        filename = self._metadata_file + ".json"
        schema_path = resources_dir / "templates" / version / "schema" / filename
        if not schema_path.exists():
            raise ValueError("Reference schema not found!")

        with open(schema_path) as file:
            schema_details = json.load(file)

        schema = {"type": schema_details.get("type"), "properties": {}, "required": schema_details.get("required")}
        for element, value in schema_details["properties"].items():
            data_type = value.get("type")
            schema["properties"][element] = {
                "type": data_type
            }

        self._schema_ref = schema

    def _execute(self, data):
        """
        Run the validation
        :param data: target data
        :type data: dict
        :return: validation status
        :rtype: bool
        """
        print("Target instance: " + str(data))
        # validate(instance=instance, schema=self._schema_ref)
        try:
            validate(instance=data, schema=self._schema_ref)
            print("Validation: Passed")
            return True
        except ValidationError:
            print(best_match(Draft202012Validator(self._schema_ref).iter_errors(data)).message)
            print("Validation: Failed")
            return False
        except Exception as e:
            print(str(e))
            print("Validation: Failed")
            return False


class Schema(object):
    def __init__(self):
        self._version = None
        self._schema = dict()
        self._schemas = dict()
        self._metadata_files = list()

        self._schema_dir = Path()

        self._column_based = ["dataset_description", "code_description"]

    @staticmethod
    def get_default_schema(version, metadata_file):
        version = convert_version_format(version)
        version = "version_" + version
        filename = metadata_file + ".json"
        schema_path = resources_dir / "templates" / version / "schema" / filename
        if not schema_path.exists():
            raise ValueError("Reference schema not found!")

        with open(schema_path) as file:
            schema = json.load(file)

        return schema

    def get_schema(self, metadata_file, version="2.0.0", print_schema=True, required_only=True, name_only=True):
        """
        get a schema via metadata_file/metadate file name

        :param metadata_file: the metadata file name
        :type metadata_file: str
        :param version: "2.0.0"|"1.2.3"
        :type version: str
        :return: dict
        """
        metadata_file = validate_metadata_file(metadata_file, version)
        filename = metadata_file + ".json"
        if version == "2.0.0":
            schema_path = resources_dir / "templates" / "version_2_0_0" / "schema" / filename
        elif version == "1.2.3":
            schema_path = resources_dir / "templates" / "version_1_2_3" / "schema" / filename
        else:
            msg = "Please provide a correct version 2.0.0 or 1.2.3"
            raise ValueError(msg)

        with open(schema_path, 'r') as file:
            schema_json: Dict = json.load(file)
            if print_schema:
                if required_only:
                    if name_only:
                        print(f"The required elements for {metadata_file}:")
                        print(json.dumps(schema_json.get('required'), indent=4))
                    else:
                        required_items = []
                        for key, value in schema_json.get('properties').items():
                            if "required" in value and value["required"] == "Y":
                                required_items.append({key: value})

                        print(f"The required elements for {metadata_file}:")
                        print(json.dumps(required_items, indent=4))
                        return required_items
                else:
                    print(json.dumps(schema_json.get('properties'), indent=4))

            return CaseInsensitiveDict(schema_json.get('properties'))

    def set_schema(self, schema):
        self._schema = schema

    def add_property(self, property_name, property):
        self._schema["properties"][property_name] = property
        required = property.get("required")
        if required == 'Y':
            self._schema["required"].append(property_name)

    def load_data(self, path):
        """
        Load in a metadata
        :param path: path to the metadata file
        :type path: string
        :return: data instance
        :rtype: dict or list
        """
        path = Path(path)
        filename = path.stem
        try:
            data_pd = pd.read_excel(path)
        except XLRDError:
            data_pd = pd.read_excel(path, engine='openpyxl')

        data_dict = data_pd.to_dict()
        return self.generate_validate_data(filename, data_dict)

    def generate_validate_data(self, filename, data_dict):
        if filename in self._column_based:
            data = {}
            elements = data_dict.get("Metadata element")
            values = data_dict.get("Value")
            length = len(elements)
            for idx in range(length):
                element = elements[idx]
                value = values[idx]
                try:
                    if math.isnan(value):
                        continue
                except:
                    pass

                data[element] = value
        else:
            data = list()

            elements = list(data_dict)
            length = len(data_dict[elements[0]])
            for idx in range(length):
                instance = {}
                for element in elements:
                    value = data_dict[element][idx]

                    try:
                        if math.isnan(value):
                            continue
                    except:
                        pass

                    instance[element] = value

                if instance:
                    data.append(instance)

        return data

    def generate_from_template(self, file, save_dir):
        """
        Generate schema from template
        (the element_description.xlsx file in the template folder)
        :param file: path to the template
        :type file: string
        :param save_dir: directory to save the schema files
        :type save_dir: string
        :return:
        :rtype:
        """
        wb = openpyxl.load_workbook(file)
        sheetnames = wb.sheetnames

        for sheet in sheetnames:
            try:
                schema_pd = pd.read_excel(file, sheet_name=sheet)
            except XLRDError:
                schema_pd = pd.read_excel(file, sheet_name=sheet, engine='openpyxl')

            schema = {"type": "object", "properties": {}, "required": []}
            required_list = list()

            for index, row in schema_pd.iterrows():
                element = row.get("Element")
                required = row.get("Required")
                data_type = row.get("Type")
                description = row.get("Description")
                example = row.get("Example")

                if required == 'Y':
                    required_list.append(element)

                # data_type = type(example).__name__

                try:
                    if math.isnan(example):
                        example = None
                    if math.isnan(description):
                        description = None
                except:
                    pass

                schema["properties"][element] = {
                    "type": data_type,
                    "required": required,
                    "description": description,
                    "example": example
                }

            schema["required"] = required_list
            self._schemas[sheet] = schema

            self.save(save_dir, schema, sheet)

    def convert_schema_excel_to_json(self, source_path, dest_path):
        """
        :param source_path: The Excel schema file path
        :type source_path: str
        :param dest_path: The converted Json file store path
        :type dest_path: str
        """
        wb = openpyxl.load_workbook(source_path)
        sheets = wb.sheetnames

        schema = dict()
        for sheet in sheets:
            schema[sheet] = dict()
            try:
                element_description = pd.read_excel(source_path, sheet_name=sheet)
            except XLRDError:
                element_description = pd.read_excel(source_path, sheet_name=sheet, engine='openpyxl')

            element_description = element_description.where(pd.notnull(element_description), None)

            for index, row in element_description.iterrows():
                element = row["Element"]
                schema[sheet][element] = dict()
                schema[sheet][element]["Required"] = row["Required"]
                schema[sheet][element]["Type"] = row["Type"]
                schema[sheet][element]["Description"] = row["Description"]
                schema[sheet][element]["Example"] = row["Example"]

        with open(dest_path, 'w') as f:
            json.dump(schema, f, indent=4)

    def save(self, save_dir, schema, metadata_file):
        """
        Save schema
        :param save_dir: path to the destination directory
        :type save_dir: string
        :param schema: metadata schema
        :type schema: dict
        :param metadata_file: metadata metadata_file (filename)
        :type metadata_file: string
        :return:
        :rtype:
        """
        save_dir = Path(save_dir)
        if not save_dir.exists():
            os.makedirs(save_dir)

        filename = '{sheet}.json'.format(sheet=metadata_file)

        save_path = Path.joinpath(save_dir, filename)
        with open(save_path, 'w', encoding='utf-8') as f:
            json.dump(schema, f, indent=4)


if __name__ == '__main__':
    # schema_xlsm = Path.joinpath(current_dir, "../resources/templates/version_1_2_3/schema.xlsx")
    # save_dir = Path.joinpath(current_dir, "../resources/templates/version_1_2_3/schema")
    schema_xlsm = Path.joinpath(current_dir, "../resources/templates/version_2_0_0/schema.xlsx")
    save_dir = Path.joinpath(current_dir, "../resources/templates/version_2_0_0/schema")

    schema = Schema()
    schema.generate_from_template(schema_xlsm, save_dir=save_dir)
    schema.convert_schema_excel_to_json(schema_xlsm, "../resources/templates/version_2_0_0/schema.json")

    # instance = {"Name": "test", "Description": "test", "Keywords":"sassf"}
    # validator = Validator()
    # validator.validate(instance, metadata_file="dataset_description", version="1.2.3")



import json
import requests
import ast


def get_api_token(cred_file, url):
    with open(cred_file) as f:
        credentials = json.load(f)
    key = {
        "api_key": credentials.get("api_key"),
        "key_id": credentials.get("key_id")
    }
    token = requests.post(url + "/user/credentials/cdis/access_token", json=key).json()

    return token


def query_data(endpoint, headers, query):
    url = endpoint + "/api/v0/submission/graphql/"

    ql = requests.post(url, json=query, headers=headers)
    data = ast.literal_eval(str(ql.text))
    return data


if __name__ == '__main__':
    experiment_submitter_ids = ["dataset-141-version-3"]
    cred_file = r"C:\Users\clin864\Desktop\sparc-codathon\sparc-me\sparc_me\gen3\credentials.json"
    endpoint = "http://gen3.abi-ctt-ctp.cloud.edu.au"

    query = {"query": """{
            project(project_id: "demo1-12L"){
                id,
                project_id,
                experiments(first: 1000){
                    id,
                    submitter_id,
                    dataset_descriptions(first: 1000){
                        id,
                        submitter_id
                    },
                    cases(first: 1000){
                        id,
                        submitter_id
                    },
                    manifests(first: 1000){
                        id,
                        submitter_id
                    }
                }
            }
        }"""}

    token = get_api_token(cred_file=cred_file, url=endpoint)
    headers = {'Authorization': 'bearer ' + token['access_token']}

    data = query_data(endpoint, headers, query)
    print(data)

    projects = data.get("data").get("project")

    experiment_uids = list()
    dataset_description_uids = list()
    case_uids = list()
    manifest_uids = list()
    for project in projects:
        experiments = project.get("experiments")
        for experiment in experiments:
            experiment_submitter_id = experiment.get("submitter_id")
            if experiment_submitter_id not in experiment_submitter_ids:
                continue
            experiment_uid = experiment.get("id")
            experiment_uids.append(experiment_uid)

            dataset_descriptions = experiment.get("dataset_descriptions")
            for dataset_description in dataset_descriptions:
                dataset_description_uid = dataset_description.get("id")
                dataset_description_uids.append(dataset_description_uid)

            cases = experiment.get("cases")
            for case in cases:
                case_uid = case.get("id")
                case_uids.append(case_uid)

            manifests = experiment.get("manifests")
            for manifest in manifests:
                manifest_uid = manifest.get("id")
                manifest_uids.append(manifest_uid)

    print("experiment_uids UIDs: " + str(experiment_uids))
    print("dataset_description UIDs: " + str(dataset_description_uids))
    print("case(subject) UIDs: " + str(case_uids))
    print("manifest UIDs: " + str(manifest_uids))

    uids = manifest_uids + case_uids + dataset_description_uids + experiment_uids
    print("UIDs: " + str(uids))



import json
from pathlib import Path
import re

metadata_files_2_0_0 = ['code_description', 'code_parameters', 'dataset_description', 'manifest', 'performances',
                        'resources',
                        'samples', 'subjects', 'submission']


def check_row_exist(dataframe, unique_column, unique_value):
    """Check if a row exist with given unique value

    :param dataframe: metadata dataframe that must be checked
    :type dataframe: Pandas DataFrame
    :param unique_value: value that can be used to uniquely identifies a row
    :type unique_value: string
    :return: row index of the row identified with the unique value, or -1 if there is no row corresponding to the unique value
    :rtype: int
    :raises ValueError: if more than one row can be identified with given unique value
    """
    row_index = dataframe.index[dataframe[unique_column] == unique_value].tolist()
    if not row_index:
        row_index = -1
    elif len(row_index) > 1:
        error_msg = "More than one row can be identified with given unique value"
        raise ValueError(error_msg)
    else:
        row_index = row_index[0]
    return row_index


def get_sub_folder_paths_in_folder(folder_path):
    """
    get sub folder paths in a folder
    :param folder_path: the parent folder path
    :type folder_path: str
    :return: list
    """
    folder = Path(folder_path)
    sub_folders = []
    for item in folder.iterdir():
        if item.is_dir():
            sub_folders.append(item)

    return sub_folders


def validate_sub_sam_name(validate_str, validate_type):
    v_4 = validate_str[:4]
    if validate_type == "sub":
        v_str = "sub-"
    elif validate_type == "sam":
        v_str = "sam-"
    else:
        error_msg = f"The validate_type should be 'sub' or 'sam', you provide validate_type is {validate_type}"
        raise ValueError(error_msg)
    if v_4 == v_str:
        return validate_str
    else:
        return f"{v_str}{validate_str}"


def validate_metadata_file(metadata_file, version="2.0.0"):
    if version == "2.0.0" or version == "2_0_0":
        for item in metadata_files_2_0_0:
            if remove_spaces_and_lower(item) == remove_spaces_and_lower(metadata_file):
                return item

        msg = f"no metadata files match {metadata_file}, please provide a correct one!"
        raise ValueError(msg)
    elif version == "1.2.3" or version == "1_2_3":
        return metadata_file


def find_col_element(element, metadata):
    elements = metadata.data.columns.tolist()
    matching_indices = metadata.validate_input(element, elements)
    if len(matching_indices) == 1:
        return elements[matching_indices[0]]
    else:
        msg = f"No '{element}' valid element is found! Please find correct element in {metadata.metadata_file}.xlsx file."
        raise KeyError(msg)


def remove_spaces_and_lower(s):
    """
    :param s: the str need to deal with
    :type s: str
    :return: a str with lower case and without space, and _
    """

    return re.sub(r'[\s_]', '', s).lower()


# Customise dict

class CaseInsensitiveDict(dict):
    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveDict, self).__init__(*args, **kwargs)
        self._key_map = {key.lower().replace(" ", ""): key for key in self.keys()}

    def __getitem__(self, key):
        original_key = self._key_map.get(key.lower().replace(" ", ""))
        if original_key is not None:
            return super(CaseInsensitiveDict, self).__getitem__(original_key)
        else:
            raise KeyError(key)

    def get(self, key, default=None, pprint=True):
        try:
            if pprint:
                print(json.dumps(self[key], indent=4))
            return self[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        self._key_map[key.lower().replace(" ", "")] = key
        super(CaseInsensitiveDict, self).__setitem__(key, value)

    def __delitem__(self, key):
        original_key = self._key_map.get(key.lower().replace(" ", ""))
        if original_key is not None:
            super(CaseInsensitiveDict, self).__delitem__(original_key)
            del self._key_map[key.lower().replace(" ", "")]



from sparc_me import Dataset

if __name__ == "__main__":
    dataset = Dataset()

    # List elements/fields
    elements = dataset.list_elements(category="dataset_description", version="1.2.3")
    # elements = dataset.list_elements(category="subjects", version="2.0.0")
    print(elements)



import requests
import json

if __name__ == '__main__':
    dataset_id = 264
    url = "https://scicrunch.org/api/1/elastic/SPARC_PortalDatasets_pr/_search"
    api_key = "xBOrIfnZTvJQtobGo8XHRvThdMYGTxtf"
    params = {'api_key': api_key}
    body = {
        "size": 10,
        "from": 0,
        "query": {
            "match": {
                "pennsieve.identifier.aggregate": {
                    "query": str(dataset_id)
                }
            }
        },
        "_source": [
            "organisms.subject.species",
            "anatomy.organ"
        ]
    }

    headers = {
        'Content-Type': 'text/plain',
    }

    response = requests.request("POST", url, headers=headers, json=body, params=params)
    result = response.json()

    print(json.dumps(result, indent=4))

    # get organs
    organ_list = list()
    hits = result.get("hits").get("hits")
    for hit in hits:
        try:
            organs = hit.get("_source").get("anatomy").get("organ")
        except AttributeError:
            continue
        for organ in organs:
            organ = organ.get("name")
            if organ:
                organ_list.append(organ)
    # remove duplicates
    organs = [*set(organ_list)]
    print("Organs:")
    print(organ_list)

    # get species
    species_list = list()
    for hit in hits:
        try:
            subjects = hit.get("_source").get("organisms").get("subject")
        except AttributeError:
            continue
        for subject in subjects:
            try:
                species = subject.get("species").get("name")
            except AttributeError:
                continue
            if species:
                species_list.append(species)
    species_list = [*set(species_list)]
    print("Species")
    print(species_list)



from pathlib import Path
from sparc_me.gen3.convertor import Gen3Convertor

if __name__ == '__main__':
    # Manifest
    experiment = "dataset-264-version-1"
    category = "manifest"
    source_dir = Path(r"C:\Users\clin864\ctt_group\sparc_datasets")
    source_dir = source_dir.joinpath(experiment)
    source = source_dir.joinpath(category + ".xlsx")
    dest_dir = source_dir

    data = {
        "type": category,
        "experiments": [{"submitter_id": experiment}],
        "submitter_id": experiment + '-' + category
    }
    convertor = Gen3Convertor()
    convertor.set_schema_dir(Path(r"C:\Users\clin864\Desktop\sparc-codathon\sparc-me\my-scripts\gen3\sds_dictionary"))
    convertor.execute(source, dest_dir=dest_dir, category=category, version="1.2.3", data_template=data)



import os
from pathlib import Path
import pandas as pd
from xlrd import XLRDError
import json
import yaml
import re


class Gen3Convertor(object):
    def __init__(self):
        self._category = None
        self._supported_versions = ["1.2.3", "2.0.0"]
        self._current_dir = Path(__file__).parent.resolve()
        self._resources_dir = self._current_dir.parent / "resources"
        self._templates_dir = self._resources_dir / "templates"
        self._schema_dir = None
        self._special_chars = ['/', '_']

        self._row_based = ["manifest", "subjects", "samples"]
        self._col_based = ["dataset_description"]

    def set_schema_dir(self, path):
        self._schema_dir = Path(path)

    def set_category(self, category):
        self._category = category

    def execute(self, source, dest_dir, category, version, data_template=dict()):
        self._validate_version(version)

        self._category = category
        mappings = self._get_mappings(version)

        data = None
        if category in self._row_based:
            if category == "manifest":
                sources = self._get_files(source, category)
            else:
                sources = [source]

            metadata_pd_composite = None
            for source in sources:
                metadata_pd = self._get_sparc_metadata(source)
                metadata_pd = self._map_fields(metadata_pd, mappings, target_version="2.0.0")

                if metadata_pd_composite is not None:
                    metadata_pd_composite = pd.concat([metadata_pd_composite, metadata_pd], ignore_index=True)
                else:
                    metadata_pd_composite = metadata_pd

            metadata_list_dict = metadata_pd_composite.to_dict('records')

            schema = self._get_schema(self._category)
            properties = schema.get("properties")
            system_properties = schema.get("systemProperties")
            required = schema.get("required")

            data_list = list()
            for metadata_dict in metadata_list_dict:
                data = data_template.copy()
                for property in properties.items():
                    key = property[0]

                    if key in system_properties:
                        continue
                    if data.get(key):
                        continue

                    value = metadata_dict.get(key)

                    # Update submitter_id
                    if key in ["subject_id", "filename"]:
                        data["submitter_id"] = data["submitter_id"] + '-' + value
                        data["submitter_id"] = data["submitter_id"].replace('./', '')
                        data["submitter_id"] = re.sub(str(self._special_chars), '-', data["submitter_id"])

                    # check if value exists and if value equals to nan (nan variable does not equal to itself)
                    if value and value == value:
                        if len(value) == 1:
                            value = value[0]

                        # handle special values
                        if isinstance(value, str):
                            value = value.replace("\"", "\'")

                            # separate string by new line "\n" and saved in a list if the list > 1
                            value_list = value.split("\n")
                            if len(value_list) > 1:
                                value = value_list

                        data[key] = value
                    else:
                        if key in required:
                            data[key] = "NA"

                data_list.append(data)
                del data
            data = data_list
        elif category in self._col_based:
            metadata_pd = self._get_sparc_metadata(source)
            metadata_pd = self._map_fields(metadata_pd, mappings, target_version="2.0.0")
            data = self._convert(metadata_pd, data_template)

        if dest_dir:
            filename = category + ".json"
            dest = dest_dir.joinpath(filename)
            self._save(data, dest)

        return data

    def _get_files(self, source, category):
        files = list()
        if source.is_file():
            files.append(source)
        elif source.is_dir():
            filename = category + ".xlsx"
            for file in source.rglob(filename):
                files.append(file)
        else:
            raise FileNotFoundError("File not found")
        return files

    def _validate_version(self, version):
        if version not in self._supported_versions:
            raise Exception("Dataset version not supported")

    @staticmethod
    def read_excel(path, sheet_name=None):
        try:
            # the read_excel method return dict when sheet name is passed. otherwise a dataframe will be returned
            if sheet_name:
                metadata = pd.read_excel(path, sheet_name=sheet_name)
            else:
                metadata = pd.read_excel(path)
        except XLRDError:
            if sheet_name:
                metadata = pd.read_excel(path, sheet_name=sheet_name, engine='openpyxl')
            else:
                metadata = pd.read_excel(path, engine='openpyxl')

        return metadata

    def _get_mappings(self, version):
        version = version.replace(".", "_")
        version = "version_" + version
        version_dir = self._templates_dir / version
        mapping_file = version_dir / "element_mapping_gen3.xlsx"

        mappings = self.read_excel(mapping_file, self._category)

        return mappings

    def _save(self, data, dest):
        os.makedirs(dest.parent, exist_ok=True)
        with open(dest, 'w') as f:
            json.dump(data, f, indent=4)
        print("Saved to " + str(dest))

    def _get_schema(self, category):
        if category == "subjects":
            category = "case"
        if category == "samples":
            category = "sample"
        schema_file = category + ".yaml"
        schema_file = self._schema_dir / schema_file
        with open(schema_file, 'r') as stream:
            schema = yaml.safe_load(stream)

            # encoding = "utf8"

        return schema

    def _get_sparc_metadata(self, source):
        metadata_sparc = self.read_excel(source)
        if source.stem == "dataset_description":
            # combine values
            metadata_sparc["values"] = metadata_sparc.iloc[:, 3:].values.tolist()
            metadata_sparc["values"] = metadata_sparc["values"].apply(lambda x: [i for i in x if str(i) != "nan"])

            # Extract columns
            # Element name: 1st column
            # Values: last column
            metadata_sparc = metadata_sparc[[metadata_sparc.columns[0], metadata_sparc.columns[-1]]]

        return metadata_sparc

    def _map_fields(self, metadata_sparc, mappings, target_version):
        if self._category == "dataset_description":
            nums_of_records = len(metadata_sparc)
            column_idx = 0
            elements = metadata_sparc.iloc[:, column_idx].tolist()

            for idx in range(nums_of_records):
                element = elements[idx]
                try:
                    record = mappings[mappings['element'] == element]
                    metadata_sparc.loc[idx, metadata_sparc.columns[0]] = record.iloc[0]["gen3_element_" + target_version]
                except:
                    continue

            metadata_sparc = metadata_sparc.dropna()
        else:
            column_headers = list(metadata_sparc.columns)
            for idx, column_header in enumerate(column_headers):
                record = mappings[mappings['element'].str.lower() == column_header.lower()]
                if record.empty:
                    continue
                new_name = record.iloc[0]["gen3_element_" + target_version]

                metadata_sparc = metadata_sparc.rename(columns={column_header: new_name})

            metadata_sparc = metadata_sparc.loc[:, metadata_sparc.columns.notna()]

        return metadata_sparc

    def _convert(self, metadata_pd, data):
        metadata_dict = dict()
        records = metadata_pd.to_dict(orient='records')

        if self._category == "dataset_description":
            for record in records:
                element = record.get("Metadata element")
                values = record.get("values")
                metadata_dict[element] = values

        schema = self._get_schema(self._category)
        properties = schema.get("properties")
        system_properties = schema.get("systemProperties")
        required = schema.get("required")

        for property in properties.items():
            key = property[0]

            if key in system_properties:
                continue
            if data.get(key):
                continue

            value = metadata_dict.get(key)

            if value:
                if self._category in self._row_based and len(value) == 1:
                    value = value[0]

                data[key] = value
            else:
                if key in required:
                    if self._category in self._col_based:
                        data[key] = []
                    else:
                        data[key] = "NA"
        return data



import json
from gen3.auth import Gen3Auth
from gen3.submission import Gen3Submission

def main():
    endpoint = "http://gen3.abi-ctt-ctp.cloud.edu.au"
    cred_file = "credentials.json"
    filename = r"C:\Users\clin864\ctt_group\sparc_datasets\dataset-46-version-2\manifest_tmp.json"
    with open(filename) as f:
        data = json.load(f)


    auth = Gen3Auth(refresh_file=cred_file)

    sub = Gen3Submission(endpoint, auth)
    sub.submit_record("demo1", "12L", json=data)


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

# ingore pybids because conflicts with digitaltwins gen3 latest version. TODO. check again later

setup(
    name="sparc_me",
    version="2.2.8",
    description='A python tool to explore, enhance, and expand SPARC datasets and their descriptions in accordance with FAIR principles.',
    author="Thiranja Prasad Babarenda Gamage, Chinchien Lin, Savindi Wijenayaka, Michael Hoffman, Linkun Gao, Haribalan Kumar",
    email="psam012@aucklanduni.ac.nz, clin864@aucklanduni.ac.nz",
    license="Apache-2.0",
    packages=find_packages(exclude=["pybids"]),
    include_package_data=True,
    package_data={'': ['resources./*']},
    install_requires=[
        'pandas',
        'styleframe',
        'xlrd',
        'openpyxl',
        'jsonschema',
        'requests'
    ]
)



import requests
import json
from pathlib import Path
import io
import pandas as pd
import queue
import threading
import os


class Dataset_Api:

    def __init__(self):
        pass

    def get_dataset_versions_pensieve(self, datasetId):
        """
        get one dataset all versions
        
        :param datasetId: the dataset id from SPARC
        :type datasetId: str|int
        :return: versions
        """

        if not isinstance(datasetId, str):
            datasetId = str(datasetId)

        url = "https://api.pennsieve.io/discover/datasets/" + datasetId + "/versions"

        headers = {"Accept": "application/json"}

        response = requests.get(url, headers=headers)

        if response.status_code == 200:
            versions = json.loads(response.text)
            return versions

    def get_all_datasets_all_versions(self):
        """
        
        Get all datasets with all versions
        It may cost a few minutes to get the whole data,
        Because some dataset have a lot of versions, e.g, 20,
        And every time when the version number getter than 1,
        it will request server for getting new data, so it waste a lot of time.

        :return: datasets
        """
        datasets = []

        latest_datasets = self.get_all_datasets_latest_version_pensieve()
        for dataset in latest_datasets:
            if dataset["version"] > 1:
                versions = self.get_dataset_versions_pensieve(dataset["id"])
                for version in versions:
                    datasets.append(version)
            else:
                datasets.append(dataset)

        return datasets

    def get_all_datasets_latest_version_pensieve(self):
        """
        Get all datasets with latest version
            
        :return: datasets | []
        """

        url = "https://api.pennsieve.io/discover/datasets?limit=2147483647&offset=0&orderBy=relevance&orderDirection=desc"

        headers = {"Accept": "application/json"}

        try:
            response = requests.get(url, headers=headers)
            response_json = json.loads(response.text)
            datasets = response_json["datasets"]

            return datasets
        except:
            print("bad connect! 404")

        return []

    def get_dataset_latest_version_pensieve(self, datasetId):
        """
        
        :param datasetId: the dataset id from SPARC
        :type datasetId: str|int
        :return:
        """

        if isinstance(datasetId, int):
            datasetId = str(datasetId)
        elif isinstance(datasetId, str):
            pass
        else:
            return
        url = "https://api.pennsieve.io/discover/datasets/" + datasetId

        headers = {"Accept": "application/json"}

        response = requests.get(url, headers=headers)

        if response.status_code == 200:
            return json.loads(response.text)

    def get_metadata_pensieve(self, datasetId, versionId):
        """
         Get a metadata from the specific version

         :param datasetId: 
         :type datasetId: str | int
         :param versionId:
         :type versionId: str | int
         :return: metadata json format
        """

        if not isinstance(datasetId, str):
            datasetId = str(datasetId)
            versionId = str(versionId)

        url = "https://api.pennsieve.io/discover/datasets/" + datasetId + "/versions/" + versionId + "/metadata"

        headers = {"Accept": "application/json"}

        response = requests.get(url, headers=headers)

        print(isinstance(response.text, str))
        if response.status_code == 200:
            return json.loads(response.text)

    def get_dataset_latest_version_number(self, datasetId):
        if not isinstance(datasetId, str):
            datasetId = str(datasetId)
        url = "https://api.pennsieve.io/discover/datasets/" + datasetId
        headers = {"Accept": "application/json"}
        response = requests.request("GET", url, headers=headers)
        response_json = json.loads(response.text)
        if response.status_code == 200:
            versionId = str(response_json['version'])
        else:
            versionId = ""
        return versionId

    def _download_file(self, datasetId, filepath):
        """
        Download bytes files from Pennsieve
        
        :param datasetId: 
        :param filepath: file path from 
        :return: 
        """
        versionId = self.get_dataset_latest_version_number(datasetId)

        url = "https://api.pennsieve.io/zipit/discover"

        payload = {"data": {
            "paths": [filepath],
            "datasetId": datasetId,
            "version": versionId
        }}
        headers = {"Content-Type": "application/json"}
        response = requests.request("POST", url, json=payload, headers=headers)
        if response.status_code == 200:
            return response
        else:
            return response.reason

    def get_xlsx_csv_file_pennsieve(self, datasetId, filepath, savepath):
        """
        
        store excel file locally
        :param datasetId: dataset id from SPARC
        :param filepath: dataset version from SPARC
        :param savepath: Path for save dataset
        """
            
        pathList = filepath.split('.')
        extension = pathList[1]
        fileStrList = filepath.split('/')
        i = len(fileStrList)
        filename = fileStrList[i - 1]
        relative_path = '/'
        for r in fileStrList[0:i - 1]:
            relative_path += r + "/"
        savepath = savepath + relative_path

        save_dir = Path(savepath)
        if not save_dir.is_dir():
            save_dir._mkdir(parents=True, exist_ok=False)
        response = self._download_file(datasetId, filepath)

        if extension == "xlsx":
            with io.BytesIO(response.content) as fh:
                df = pd.io.excel.read_excel(fh, engine='openpyxl')
            df.dropna(axis=0, how='all', inplace=True)
            writer = pd.ExcelWriter(savepath + filename)
            df.to_excel(writer)
            writer.save()

        elif extension == "csv":
            with io.BytesIO(response.content) as fh:
                df = pd.read_csv(fh)
            df.to_csv(savepath + filename, sep=',', header=False, index=False)

    def get_UBERONs_From_Dataset(self, datasetId, filepath):
        response = self._download_file(datasetId, filepath)
        with io.BytesIO(response.content) as fh:
            df = pd.read_csv(fh)
        df = df.dropna(axis=0, how='any')
        return df['Term ID']

    '''
    TODO: download whole dataset
    '''

    def _mkdir(self, paths):
        for path in paths:
            savepath = "dataset/"
            fileStrList = path.split('/')
            i = len(fileStrList)
            relative_path = '/'
            for r in fileStrList[0:i - 1]:
                relative_path += r + "/"
            savepath = savepath + relative_path
            folder = os.path.exists(savepath)
            if not folder:
                os.makedirs(savepath)

    def get_all_files_path(self, dataset_id, version_id):
        url = f"https://api.pennsieve.io/discover/datasets/{dataset_id}/versions/{version_id}/metadata"

        headers = {"Accept": "application/json"}

        response = requests.get(url, headers=headers)

        if response.status_code == 200:
            files = json.loads(response.text)["files"]
            paths = []
            for idx in range(len(files)):
                paths.append(files[idx]["path"])
            return paths

    def _craw(self, datasetId, versionId, url_queue: queue.Queue, html_queue: queue.Queue):
        '''
          Download bytes files from Pennsieve
        '''
        while True:
            if url_queue.qsize() == 0:
                break
            filepath = url_queue.get()
            if filepath is None:
                html_queue.put(None)
                continue

            url = "https://api.pennsieve.io/zipit/discover"

            payload = {"data": {
                "paths": [filepath],
                "datasetId": datasetId,
                "version": versionId
            }}
            headers = {"Content-Type": "application/json"}
            try:
                response = requests.request("POST", url, json=payload, headers=headers, stream=True)
                if response.status_code == 200:
                    contentObj = {
                        "content": response.content,
                        "filepath": filepath,
                    }
                    html_queue.put(contentObj)
            except Exception as e:
                print(f"The file: {filepath} download failed! The error is {e}")

    def _parse(self, html_queue: queue.Queue):
        while True:
            res = html_queue.get()
            if res is None:
                print("finish downloading dataset!")
                break
            with io.BytesIO(res["content"]) as io_file:
                with open("dataset/" + res["filepath"], 'wb') as file:
                    file.write(io_file.getvalue())

    def download_dataset(self, dataset_id, version_id=None):
        latest_version_id = self.get_dataset_latest_version_number(dataset_id)
        if latest_version_id == '':
            print("No dataset from SPARC!")
            return
        if version_id is None:
            version_id = latest_version_id
        version_id = str(version_id)
        if int(version_id) > int(latest_version_id):
            version_id = latest_version_id
            print(
                "Your input version ID is greater that the one exist in SPARC dataset, Now will download the latest version in SPARC for you!")
        if int(version_id) < 1:
            version_id = "1"
            print("Invalid version id, Now will download the first version of the dataset for you!")

        paths = self.get_all_files_path(dataset_id, version_id)
        self._mkdir(paths)
        url_queue = queue.Queue()
        html_queue = queue.Queue()
        threads = []

        for url in paths:
            url_queue.put(url)

        url_queue.put(None)
        url_queue.put(None)
        url_queue.put(None)

        for idx in range(3):
            t1 = threading.Thread(target=self._craw, args=(dataset_id, version_id, url_queue, html_queue))
            threads.append(t1)
            t1.start()
        for idx in range(2):
            t2 = threading.Thread(target=self._parse, args=(html_queue,))
            t2.start()

        for t in threads:
            t.join()

    def get_dataset_protocolsio_link(self, datasetId):
        dataset = self.get_dataset_latest_version_pensieve(datasetId)
        protocol_url = ""

        if dataset:
            if (len(dataset["externalPublications"]) > 0):
                protocol_url = dataset["externalPublications"][0]["doi"]

        return protocol_url

    def get_protocolsio_text(self, datasetId, dir):
        save_dir = Path(dir)
        if not save_dir.is_dir():
            save_dir._mkdir(parents=True, exist_ok=False)

        protocol_url = self.get_dataset_protocolsio_link(datasetId)
        if protocol_url:
            doi = protocol_url
            url = "https://www.protocols.io/api/v4/protocols/" + doi
            querystring = {
                "Authorization": "9335c865c035bcabc986e443e4bfab3547fd3c8a4e052746ac2a10290d91b6cb",
            }
            headers = {
                "Accept": "*/*",
                "Content-Type": "application/json"
            }
            response = requests.request(
                "GET", url, headers=headers, params=querystring)
            if response.status_code == 200:
                protocol_json = json.loads(response.content)
                with open(dir + '/protocol_data.json', 'w') as f:
                    json.dump(protocol_json, f)



import pandas as pd
from pathlib import Path
from sparc_me.core.utils import remove_spaces_and_lower
import shutil
from sparc_me.core.utils import find_col_element
from datetime import datetime, timezone
from typing import List


class Metadata:
    def __init__(self, metadata_file, metadata, version, dataset_path):
        """
        :param metadata_file: metadata file name
        :type metadata_file: str
        :param metadata: metadata dataframe content
        :type metadata: Dataframe
        :param version: dataset version
        :type version: "2.0.0" | "1.2.3"
        :param dataset_path: root dataset path
        :type dataset_path: Path
        """
        self.metadata_file = metadata_file
        self.data = metadata
        self.version = version
        self.metadata_file_path = Path(dataset_path).joinpath(f"{metadata_file}.xlsx")

    """********************************* Add values *********************************"""

    def add_values(self, element, values):
        """
        Append values to the last value

        :param element: a col/row name in Excel
        :type element: str
        :param values: element values from that metadata
        :type values: list | str | int | bool
        :return:
        """
        self._add_values(element, values, append=True)

    def set_values(self, element, values):
        """
        Override the values in that element

        :param element: a col/row name in Excel
        :type element: str
        :param values: element values from that metadata
        :type values: list | str | int | bool
        :return:
        """
        self._add_values(element, values, append=False)

    def _add_values(self, element, values, append):
        """
        Set single cell. The row is identified by the given unique name and column is identified by the header.

        :param element: a col/row name in Excel
        :type element: str
        :param values: element values from that metadata
        :type values: list | str | int | bool
        :param append: insert values into last col or row
        :type append: bool
        """
        values = self._validate_input_values(values)
        if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
            if element == '':
                return
            else:
                col_name = 'Value'
            excel_row_index = self._find_row_index(element)
            self.set_row_values(row_index=excel_row_index, values=values, col_name=col_name, append=append)
        else:
            if element == '':
                # add values by rows start at (0,0)
                # get first header
                header = self.data.columns[0]
                self.set_row_values(row_index=2, values=values, col_name=header, append=append)
            else:
                # add values by header (columns)
                col_index = self._find_col_index(element)
                self.set_col_values(col_index=col_index, values=values, append=append)

    def set_row_values(self, row_index, values, col_name='Value', append=True):
        """
        Set row fields/values by row idx/name and column name (the header)

        :param row_index: row index in Excel. Excel index starts from 1 where index 1 is the header row. so actual data index starts from 2
        :type row_index: int
        :param col_name: column name. the header is the first row
        :type col_name: str
        :param values: field values
        :type values: list | str | int | bool
        :param append: insert values into last col or row
        :type append: bool
        :return: updated dataset
        :rtype: dict
        """
        values = self._validate_input_values(values)
        insert_header = col_name
        if not isinstance(row_index, int):
            msg = "row_index should be 'int'."
            raise ValueError(msg)
        if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
            try:
                row_index = row_index - 2
                # find out column index
                row_has_null = self.data.iloc[row_index].isnull().any()
                if append:
                    if row_has_null:
                        start_header = (self.data.isnull().idxmax(axis=1))[row_index]
                        insert_header = start_header
                        start_column_index = self.data.columns.get_loc(start_header)
                    else:
                        last_value_col_index = len(self.data.iloc[row_index].values) - 1
                        last_value_col_name = self.data.columns[last_value_col_index]
                        value_header_order = last_value_col_name.replace('Value', '')
                        if value_header_order == '':
                            self.data.insert(last_value_col_index + 1, f"Value {1}", None)
                        else:
                            try:
                                if value_header_order == ' n':
                                    value_header_order = 3
                                self.data.insert(last_value_col_index + 1, f"Value {int(value_header_order) + 1}",
                                                 None)
                            except ValueError:
                                msg = "Please private a correct header, e.g, Value, Value1, Value2..."
                                raise ValueError(msg)

                        insert_header = self.data.columns[-1]
                        start_column_index = last_value_col_index + 1
                else:
                    start_column_index = self.data.columns.get_loc(col_name)

                if len(values) > 0:
                    self._edit_column(insert_header, len(values))
                    self.data.iloc[row_index, start_column_index:start_column_index + len(values)] = values
                else:
                    msg = f"please provide values"
                    raise ValueError(msg)
                # Convert Excel row index to dataframe index: index - 2

            except ValueError:
                msg = "Value error. row does not exists."
                raise ValueError(msg)
        else:
            if append:
                self.data.loc[len(self.data)] = values
            else:
                self.data.loc[-1] = values
                self.data.index = self.data.index + 1
                self.data.sort_index()

    def set_col_values(self, col_index, values, append=True):
        """
        Set col fields/values by col index
        :param values: field values
        :type values: list | str | int | bool
        :param col_index: the header index
        :type col_index: int
        :param append: insert values into last col or row
        :type append: bool
        :return:
        """
        values = self._validate_input_values(values)
        if append:
            if self.data.iloc[:, col_index].isnull().any():
                nan_row_index = self.data[self.data.iloc[:, col_index].isnull()].index[0]
                self.data.iloc[nan_row_index:nan_row_index + len(values), col_index] = values
            else:
                for value in values:
                    new_row = [None] * len(self.data.columns)
                    new_row[col_index] = value
                    self.data.loc[len(self.data)] = new_row
        else:
            if len(self.data) < len(values):
                diff = len(values) - len(self.data)
                for _ in range(diff):
                    self.data.loc[len(self.data)] = None

            for i, value in enumerate(values):
                self.data.iat[i, col_index] = value

    """********************************* Clear & Remove values *********************************"""

    def clear_values(self, element=''):
        """
        :param element:  Unique row/col name in Excel. (Ex: if subjects is metadata_file, a row name can be a unique subjet id)
        :type element: str
        :return:
        """
        if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
            header_name = 'Value'
            if element == '':
                self.data.fillna('None', inplace=True)
                self.data.drop(columns=self.data.columns[self.data.columns.get_loc(header_name):],
                               inplace=True)
                self.data['Value'] = pd.NA
                if self.metadata_file == "dataset_description":
                    self.set_values(element="Metadata version", values="2.0.0")
            else:
                excel_row_index = self._find_row_index(element)
                df_row_index = excel_row_index - 2
                header_index = self.data.columns.get_loc(header_name)
                self.data.iloc[df_row_index, header_index:] = pd.NA
        else:
            if element == '':
                self.data.drop(self.data.index, inplace=True)
            else:
                col_index = self._find_col_index(element)
                self.data.iloc[:, col_index] = pd.NA

    def remove_values(self, element, values):
        """
        :param values: field value
        :type values: list | str | int | bool
        :param element: Unique row/col name in Excel.
        :type element: str
        :return:
        """
        validate_values = self._validate_input_values(values)
        if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
            excel_row_index = self._find_row_index(element)
            self._remove_values(field_index=excel_row_index, values=validate_values)
        else:
            col_index = self._find_col_index(element)
            self._remove_values(field_index=col_index, values=validate_values)

    def remove_row(self, value):
        rows_to_delete = self.data[self.data.eq(value).any(axis=1)]
        self.data.drop(rows_to_delete.index, inplace=True)

    def _remove_values(self, field_index, values):
        """
        :param values: field values
        :type values: list | int | str | bool
        :param field_index: a col/row name index in Excel
        :type field_index: int
        :return:
        """
        # get all values from this row
        excel_field_index = field_index
        current_values = self._get_values(excel_field_index)
        df_field_index = field_index - 2
        for value in values:
            if value in current_values.tolist():
                self.data.fillna('None', inplace=True)
                if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
                    column_with_value = self.data.loc[df_field_index].eq(value)
                    self.data.loc[df_field_index, column_with_value] = 'None'
                else:
                    self.data.loc[
                        self.data.iloc[:, field_index] == value, self.data.columns[field_index]] = 'None'
        self.data[self.data == 'None'] = pd.NA

    """********************************* Get values *********************************"""

    def get_values(self, element):
        """

        :param element: a col/row name in Excel
        :return:
        """
        if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
            excel_row_index = self._find_row_index(element)
            return self._get_values(excel_row_index)
        else:
            col_index = self._find_col_index(element)
            return self._get_values(col_index)

    def _get_values(self, field_index):
        """
        :param field_index: a col/row name index in Excel
        :return:
        """
        if self.metadata_file == "dataset_description" or self.metadata_file == "code_description":
            value_header_start = self.data.columns.get_loc('Value')
            values = self.data.iloc[field_index - 2, value_header_start:]
        else:
            values = self.data.iloc[:, field_index]
        return values

    """********************************* Find col/row index *********************************"""

    def _find_row_index(self, row_name):
        """

        :param row_name: a str row name in Excel
        :type row_name: str
        :return:
        """
        elements = self.data[self.data.columns[0]].tolist()
        matching_indices = self.validate_input(row_name, elements)
        if not matching_indices:
            msg = f"No row with given unique name, {row_name}, was found in the unique column {self.data.columns[0]}"
            raise ValueError(msg)
        elif len(matching_indices) > 1:
            msg = f"More than one row with given unique name, {row_name}, was found in the unique column {self.data.columns[0]}"
            raise ValueError(msg)
        else:
            excel_row_index = matching_indices[0] + 2
            return excel_row_index

    def _find_col_index(self, col_name):
        """
        :param col_name: a str col/header name in Excel
        :type col_name: str
        :return:
        """
        elements = self.data.columns.tolist()
        matching_indices = self.validate_input(col_name, elements)
        if len(matching_indices) == 1:
            return matching_indices[0]
        else:
            msg = f"No valid field name is found!"
            raise KeyError(msg)

    """********************************* Edit column *********************************"""

    def _edit_column(self, insert_header, nums, operator='+'):
        """

        :param insert_header: the start header name
        :type insert_header: str
        :param nums: expand or reduce column's numbers
        :type nums: int
        :param operator: identifier for expand(+) or reduce(-) columns
        :type operator: "+" | "-"
        :return:
        """
        start_column_idx = self.data.columns.get_loc(insert_header)
        remain_headers = self.data.columns[start_column_idx:]
        last_header_index = self.data.columns.get_loc(self.data.columns[-1])
        if operator == "+":
            if len(remain_headers) < nums:
                length = nums - len(remain_headers)
                for i in range(length):
                    idx = i + 1
                    (idx, unique_header) = self._create_unique_header(idx)

                    try:
                        self.data.insert(last_header_index + i + 1, unique_header, None)
                    except ValueError:
                        msg = "Please provide a correct header, e.g, Value, Value 1, Value 2..."
                        raise ValueError(msg)

    def _create_unique_header(self, idx):
        while True:
            if f"Value {idx}" in self.data.columns:
                idx += 1
            else:
                return (idx, f"Value {idx}")

    """********************************* Validate inputs *********************************"""

    def validate_input(self, element, elements):
        """

        :param element: row/col name in excel
        :param elements: dataset metadata elements
        :return:
        """
        if not isinstance(element, str):
            msg = "row_name / col_name should be string."
            raise ValueError(msg)

        # Assumes that all excel files first column contains the unique value field
        # TODO: In version 1, the unique column is not the column 0. Hence, unique column must be specified.
        # This method need to be fixed to accomadate this

        row_name_cleaned = remove_spaces_and_lower(element)
        matching_indices = [index for index, value in enumerate(elements) if
                            remove_spaces_and_lower(value) == row_name_cleaned]
        return matching_indices

    def _validate_input_values(self, values):
        new_values = []
        if isinstance(values, str) or isinstance(values, int) or isinstance(values, bool):
            new_values.append(values)
        elif not isinstance(values, list):
            msg = "Please provide a correct values parameter, if is a single value, use str, int, bool, if multiple values please put values into a list."
            raise TypeError(msg)
        else:
            new_values = values
        return new_values

    """********************************* Save *********************************"""

    def save(self, path=""):
        """
        Save the metadata file to dataset dir

        :param path: the dataset root path (Optional)
        :type path: str
        """
        try:
            if path == "":
                path = self.metadata_file_path
            self.data.to_excel(path, index=False)
        except:
            msg = f"Please provide a correct path for {self.metadata_file}"
            raise ValueError(msg)


class Sample:
    count = 0
    _dataset_path: Path = Path('./')
    _metadata: Metadata = None
    _manifest_metadata: Metadata = None


    def __init__(self):
        self.sample_id = ""
        self.subject_id = ""
        self.sample_dir = Path()
        self.source_sample_paths: List[Path] = []
        self.index = -1

    def set_subject_id(self, sub_id):
        """
        set subject id to sample object
        :param sub_id: a subject id
        :type sub_id: str
        """
        self.subject_id = sub_id
        self._generate_sample_path_and_id()

    def _generate_sample_path_and_id(self):
        """
        Generate sample path and id
        """
        subject_dir = self._dataset_path.joinpath("primary", self.subject_id)
        if subject_dir.exists():
            sub_dirs = []
            for sub_dir in subject_dir.iterdir():
                if sub_dir.is_dir():
                    sub_dirs.append(sub_dir.name)
            while True:
                Sample.count += 1
                self.sample_id = f"sam-{Sample.count}"
                if self.sample_id not in sub_dirs:
                    break
        else:
            Sample.count += 1
            self.sample_id = f"sam-{Sample.count}"

        self.sample_dir = subject_dir.joinpath(self.sample_id)
        self._add_sample_row()

    def get_sample_id(self):
        """
        Get sample id
        :return: sample id
        """
        return self.sample_id

    def _add_sample_row(self):
        """
        Add row in sample metadata
        """
        df = self._metadata.data
        if self.sample_id in df['sample id'].values and self.subject_id in df['subject id']:
            self.index = df.loc[df['sample id'] == self.sample_id].index[0]
        else:
            sample = [self.sample_id, self.subject_id] + [float('nan')] * (len(df.columns) - 2)
            # Create new row
            self.index = len(df)
            df.loc[self.index] = sample

    def add_path(self, source_path):
        """
        Add sample source path to sample object

        :param source_path: sample folder source path
        :type source_path: str | list

        """
        if isinstance(source_path, list):
            for file_path in source_path:
                self.source_sample_paths.append(Path(file_path))
        else:
            self.source_sample_paths.append(Path(source_path))


    def set_path(self, source_path):
        """
        Add sample source path to sample object
        Override the Previous path

        :param source_path: sample folder source path
        :type source_path: str | list

        """
        if isinstance(source_path, list):
            self.source_sample_paths = []
            for file_path in source_path:
                self.source_sample_paths.append(Path(file_path))
        else:
            self.source_sample_paths = [Path(source_path)]


    def set_values(self, metadata={}):
        """
        :param metadata: key : value dict (element:value)
        :type metadata: dict
        """
        if not isinstance(metadata, dict):
            msg = f"You should use a dict here, you provide parameter type is {type(metadata)}"
            raise TypeError(msg)

        for element, value in metadata.items():
            if element == 'sample id' or element == 'subject id':
                continue
            else:
                self.set_value(element, value)

    def set_value(self, element, value):
        """
        Set value for one element for a sample

        :param element: element in sample metadata file
        :type element: str
        :param value: the values for that element
        :type value: str|int

        """
        df = self._metadata.data
        index = df.loc[(df['sample id'] == self.sample_id) & (df['subject id'] == self.subject_id)].index[0]
        element = find_col_element(element, self._metadata)
        if index == self.index:
            df.loc[index, element] = value

        self.save()

    def move(self):
        """
        Move sample files from source dir to dataset primary dir

        """
        if not self.sample_dir.exists():
            self.sample_dir.mkdir(parents=True, exist_ok=True)

        for source_sam in self.source_sample_paths:
            if source_sam.is_dir():
                source_sample_files = source_sam.rglob("*")
                for file in source_sample_files:
                    if file.is_file():
                        relative_path = file.relative_to(source_sam)
                        target_file = self.sample_dir / relative_path
                        target_file.parent.mkdir(parents=True, exist_ok=True)
                        shutil.copy(str(file), str(target_file))
                        self._update_manifest(sample_path=str(target_file))
            elif source_sam.is_file():
                shutil.copy(str(source_sam), str(self.sample_dir))
                self._update_manifest(sample_path=str(self.sample_dir / source_sam.name))
    def _update_manifest(self, sample_path):
        """
        Update manifest metadata, after remove samples

        :param sample_path: sample path
        :type sample_path: str
        """
        file_path = Path(
            sample_path.replace(str(self._dataset_path), '')[1:]).as_posix()

        row = {
            'timestamp': datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
            'description': f"File of subject {self.subject_id} sample {self.sample_id}",
            'file type': Path(sample_path).suffix
        }

        df_manifest = self._manifest_metadata.data
        # check is exist
        if file_path in df_manifest['filename'].values:
            manifest_index = df_manifest.loc[df_manifest['filename'] == file_path].index[0]
        else:
            manifest_row = [file_path] + [float('nan')] * (len(df_manifest.columns) - 1)
            # Create new row
            manifest_index = len(df_manifest)
            df_manifest.loc[manifest_index] = manifest_row

        for element, value in row.items():
            validate_element = find_col_element(element, self._manifest_metadata)
            df_manifest.loc[manifest_index, validate_element] = value

        self._manifest_metadata.save()

    def remove_values(self):
        """
        Future function
        """
        pass

    def save(self):
        """
        Save sample metadata file

        """
        self._metadata.save()

class Subject:
    count = 0
    _dataset_path = Path('./')
    _metadata = None

    def __init__(self):

        self.subject_id = ""
        self.subject_dir = Path()
        self.index = -1
        self._samples = {}
        self._generate_subject_path_and_id()

    def get_sample(self, sample_sds_id) -> Sample:
        """
        Provide the sample sds id to query a sample for edit

        :param sample_sds_id: the sample id in sds
        :return: Sample
        """
        if not isinstance(sample_sds_id, str):
            msg = f"Sample not found, please provide an string sample_sds_id!, you sample_sds_id type is {type(sample_sds_id)}"
            raise ValueError(msg)

        try:
            sample = self._samples.get(sample_sds_id)
            return sample
        except:
            msg = f"Sample not found with {sample_sds_id}! Please check your subject_sds_id in subject metadata file"
            raise ValueError(msg)

    def _generate_subject_path_and_id(self):
        """
        generate subject id and path in dataset

        """
        primary_dir = self._dataset_path.joinpath("primary")
        if primary_dir.exists():
            sub_dirs = []
            for sub_dir in primary_dir.iterdir():
                if sub_dir.is_dir():
                    sub_dirs.append(sub_dir.name)
            while True:
                Subject.count += 1
                self.subject_id = f"sub-{Subject.count}"
                if self.subject_id not in sub_dirs:
                    break
        else:
            Subject.count += 1
            self.subject_id = f"sub-{Subject.count}"
        self.subject_dir = primary_dir.joinpath(self.subject_id)
        self._add_subject_row()

    def add_samples(self, samples):
        """
        Add samples into subject object

        :param samples: a samples list
        :type samples: list
        """
        if isinstance(samples, list):
            for sample in samples:
                if isinstance(sample, Sample):
                    self._create_sample(sample)
                    self._samples[sample.sample_id] = sample
        elif isinstance(samples, Sample):
            self._create_sample(samples)
            self._samples[samples.sample_id] = samples

    def _create_sample(self, sample):
        """
        Generate sample id and subject id in sample class
        :param sample: a sample object
        :type sample: Sample
        """
        sample.set_subject_id(self.subject_id)

    def _add_subject_row(self):
        """
        Add row in subject metadata file

        """
        df = self._metadata.data
        if self.subject_id in df['subject id']:
            self.index = df.loc[df['subject id'] == self.subject_id].index[0]
        else:
            subject = [self.subject_id] + [float('nan')] * (len(df.columns) - 1)
            # Create new row
            self.index = len(df)
            df.loc[self.index] = subject

    def set_values(self, metadata={}):
        """
        :param metadata: key : value dict (element:value)
        :type metadata: dict
        """
        if not isinstance(metadata, dict):
            msg = f"You should use a dict here, you provide parameter type is {type(metadata)}"
            raise TypeError(msg)

        for element, value in metadata.items():
            if element == 'subject id':
                continue
            else:
                self.set_value(element, value)

    def set_value(self, element, value):
        """
        Set value for one element for a subject

        :param element: element in sample metadata file
        :type element: str
        :param value: the values for that element
        :type value: str|int
        """
        df = self._metadata.data
        index = df.loc[df['subject id'] == self.subject_id].index[0]
        element = find_col_element(element, self._metadata)
        if index == self.index:
            df.loc[index, element] = value

        self.save()

    def move(self):
        for sample in self._samples.values():
            sample.move()

    def remove_values(self):
        pass

    def save(self):
        self._metadata.save()




from pathlib import Path
import json
import os

if __name__ == '__main__':
    dest = Path(r"C:\Users\clin864\ctt_group\sparc_datasets\gen3\dataset-12L_7-version-1")
    data = {
        "type": "experiment",
        "submitter_id": "dataset-12L_7-version-1",
        "projects": [
            {
                "code": "12L"
            }
        ],
        # "experimental_description": "MRI BREAST BILATERAL WWO",
        # "data_description": "https://doi.org/10.7937/TCIA.e3svre93"
    }

    # saving
    os.makedirs(dest, exist_ok=True)

    dest = dest.joinpath("experiment.json")
    with open(dest, 'w') as f:
        json.dump(data, f, indent=4)
    print("Saved to " + str(dest))



from gen3.auth import Gen3Auth
from gen3.submission import Gen3Submission

def main():
    endpoint = "http://gen3.abi-ctt-ctp.cloud.edu.au"
    cred_file = "credentials.json"

    auth = Gen3Auth(refresh_file=cred_file)

    sub = Gen3Submission(endpoint, auth)
    program = "demo1"
    project = "12L"
    node_type = "dataset_description"
    fileformat = "json"
    filename = r"C:\Users\clin864\ctt_group\sparc_datasets\dataset_description_all_tmp.json"
    # sub.export_node(program, project, node_type, fileformat, filename)
    sub.export_node(program, project, node_type, fileformat)


if __name__ == "__main__":
    main()
